# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import math
import StOptReg as reg
import StOptGrids
import StOptGlobal
import Utils
import Simulators as sim
import Optimizers as opt
import dp.DynamicProgrammingByRegressionDist as dynmpi
import dp.SimulateRegressionControlDist as srtmpi
import unittest
import importlib

accuracyClose = 1e6
accuracyEqual = 0.0001

# valorization of a given gas storage on a  grid
# p_grid             the grid
# p_maxLevelStorage  maximum level
def gasStorage(p_grid, p_maxLevelStorage) :
    
    from mpi4py import MPI
    world = MPI.COMM_WORLD
    # storage
    injectionRateStorage = 60000.
    withdrawalRateStorage = 45000.
    injectionCostStorage = 0.35
    withdrawalCostStorage = 0.35

    maturity = 1.
    nstep = 10

    # define a a time grid
    timeGrid = StOptGrids.OneDimRegularSpaceGrid(0., maturity / nstep, nstep)
    # future values
    futValues = []

    # periodicity factor
    iPeriod = 52

    for i in list(range(nstep + 1)) :
        futValues.append(50. + 20. * math.sin((math.pi * i * iPeriod) / nstep))

    # define the future curve
    futureGrid = Utils.FutureCurve(timeGrid, futValues)

    # one dimensional factors
    nDim = 1
    sigma = np.zeros(nDim) + 0.94
    mr = np.zeros(nDim) + 0.29
    # number of simulations
    nbsimulOpt = 20000

    # no actualization
    rate =0.
    # a backward simulator
    bForward = False

    backSimulator = sim.MeanRevertingSimulator(futureGrid, sigma, mr, rate, maturity, nstep, nbsimulOpt, bForward)
    # optimizer                  
    storage = opt.OptimizeGasStorageMeanReverting(injectionRateStorage, withdrawalRateStorage, injectionCostStorage, withdrawalCostStorage)

    # regressor
    nMesh = 6
    nbMesh = np.zeros(1, dtype = np.int32) + nMesh
    regressor = reg.LocalLinearRegression(nbMesh)
    # final value
    vFunction = Utils.ZeroPayOff()

    # initial values
    initialStock = np.zeros(1) + p_maxLevelStorage
    initialRegime = 0 # only one regime

    # Optimize
    fileToDump = "CondExpGasStorageHLMpi"
    bOneFile = True

    # link the simulations to the optimizer
    storage.setSimulator(backSimulator)
    valueOptimMpi = dynmpi.DynamicProgrammingByRegressionDist(p_grid, storage, regressor, vFunction, initialStock, initialRegime, fileToDump, bOneFile)
    print("valOptimMpi", valueOptimMpi)

    world.barrier()

    nbsimulSim = 40000
    bForward = True
    forSimulator = sim.MeanRevertingSimulator(futureGrid, sigma, mr, rate,maturity, nstep, nbsimulSim, bForward)
    storage.setSimulator(forSimulator)
    valSimuMpi = srtmpi.SimulateRegressionControlDist(p_grid, storage, vFunction, initialStock, initialRegime, fileToDump, bOneFile)
    print("valSimuMpi", valSimuMpi)

    return valueOptimMpi, valSimuMpi
                    
class testGasStorageTest(unittest.TestCase):
           
    def test_simpleStorageMpi(self):
               
        moduleMpi4Py=importlib.util.find_spec('mpi4py')
        if (moduleMpi4Py is not None):
            from mpi4py import MPI
            world = MPI.COMM_WORLD
            # storage
            maxLevelStorage = 90000
            # grid
            nGrid = 10
            lowValues = np.zeros(1)
            step = np.zeros(1) + (maxLevelStorage / nGrid)
            nbStep = np.zeros(1, dtype = np.int32) + nGrid
            grid = StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)
                   
            val = gasStorage(grid, maxLevelStorage)
            
            if world.rank == 0:
                self.assertAlmostEqual(val[0], val[1], None, None, accuracyClose)    
            
            world.barrier()
            
            # grid
            ######
            poly = np.zeros(1, dtype = np.int32) + 1
            gridL = StOptGrids.RegularLegendreGrid(lowValues, step, nbStep, poly)
            
            valLegendre = gasStorage(gridL, maxLevelStorage)
            
            if world.rank == 0:
                self.assertAlmostEqual(valLegendre[0], valLegendre[1], None, None, accuracyClose)
         
    def test_simpleStorageLegendreQuadratic(self):
         
        moduleMpi4Py=importlib.util.find_spec('mpi4py')
        if (moduleMpi4Py is not None):
            from mpi4py import MPI
            world = MPI.COMM_WORLD
            # storage
            maxLevelStorage = 90000
            # grid
            nGrid = 5
            lowValues = np.zeros(1)
            step = np.zeros(1) + (maxLevelStorage / nGrid)
            nbStep = np.zeros(1, dtype = np.int32) + nGrid
            poly = np.zeros(1, dtype = np.int32) + 2
            grid = StOptGrids.RegularLegendreGrid(lowValues, step, nbStep, poly)
             
            val = gasStorage(grid, maxLevelStorage)
            
            if world.rank == 0:
                self.assertAlmostEqual(val[0], val[1], None, None, accuracyClose)
         
    def test_simpleStorageLegendreCubic(self):
         
        moduleMpi4Py=importlib.util.find_spec('mpi4py')
        if (moduleMpi4Py is not None):
            from mpi4py import MPI
            world = MPI.COMM_WORLD
            # storage
            maxLevelStorage = 90000
            # grid
            nGrid = 5
            lowValues = np.zeros(1)
            step = np.zeros(1) + (maxLevelStorage / nGrid)
            nbStep = np.zeros(1, dtype = np.int32) + nGrid
            poly = np.zeros(1, dtype = np.int32) + 3
            grid = StOptGrids.RegularLegendreGrid(lowValues, step, nbStep, poly)
             
            val = gasStorage(grid, maxLevelStorage)
            
            if world.rank == 0:
                self.assertAlmostEqual(val[0], val[1], None, None, accuracyClose)
              
if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python

import math
from numpy import *

def i4_bit_hi1 ( n ):

#*****************************************************************************80
#
## I4_BIT_HI1 returns the position of the high 1 bit base 2 in an integer.
#
#  Example:
#
#       N    Binary     BIT
#    ----    --------  ----
#       0           0     0
#       1           1     1
#       2          10     2
#       3          11     2 
#       4         100     3
#       5         101     3
#       6         110     3
#       7         111     3
#       8        1000     4
#       9        1001     4
#      10        1010     4
#      11        1011     4
#      12        1100     4
#      13        1101     4
#      14        1110     4
#      15        1111     4
#      16       10000     5
#      17       10001     5
#    1023  1111111111    10
#    1024 10000000000    11
#    1025 10000000001    11
#
#  Licensing:
#
#    This code is distributed under the MIT license.
#
#  Modified:
#
#    22 February 2011
#
#  Author:
#
#    Original MATLAB version by John Burkardt.
#    PYTHON version by Corrado Chisari
#
#  Parameters:
#
#    Input, integer N, the integer to be measured.
#    N should be nonnegative.  If N is nonpositive, the value will always be 0.
#
#    Output, integer BIT, the number of bits base 2.
#
  i = int ( n )
  bit = 0
  while ( True ):
    if ( i <= 0 ):
      break
    bit += 1
    i = ( i // 2 )
  return bit

def i4_bit_lo0 ( n ):

#*****************************************************************************80
#
## I4_BIT_LO0 returns the position of the low 0 bit base 2 in an integer.
#
#  Example:
#
#       N    Binary     BIT
#    ----    --------  ----
#       0           0     1
#       1           1     2
#       2          10     1
#       3          11     3 
#       4         100     1
#       5         101     2
#       6         110     1
#       7         111     4
#       8        1000     1
#       9        1001     2
#      10        1010     1
#      11        1011     3
#      12        1100     1
#      13        1101     2
#      14        1110     1
#      15        1111     5
#      16       10000     1
#      17       10001     2
#    1023  1111111111     1
#    1024 10000000000     1
#    1025 10000000001     1
#
#  Licensing:
#
#    This code is distributed under the MIT license.
#
#  Modified:
#
#    22 February 2011
#
#  Author:
#
#    Original MATLAB version by John Burkardt.
#    Python version by Corrado Chisari
#
#  Parameters:
#
#    Input, integer N, the integer to be measured.
#    N should be nonnegative.
#
#    Output, integer BIT, the position of the low 1 bit.
#
  bit = 0
  i = int ( n )
  while ( 1 ):
    bit = bit + 1
    i2 = ( i // 2 )
    if ( i == 2 * i2 ):
      break
    i = i2

  return bit
	
def i4_sobol_generate ( m, n, skip ):

#*****************************************************************************80
#
## I4_SOBOL_GENERATE generates a Sobol dataset.
#
#  Licensing:
#
#    This code is distributed under the MIT license.
#
#  Modified:
#
#    22 February 2011
#
#  Author:
#
#    Original MATLAB version by John Burkardt.
#    PYTHON version by Corrado Chisari
#
#  Parameters:
#
#    Input, integer M, the spatial dimension.
#
#    Input, integer N, the number of points to generate.
#
#    Input, integer SKIP, the number of initial points to skip.
#
#    Output, real R(M,N), the points.
#
	r=zeros((m,n))
	for j in range (1, n+1):
		seed = skip + j - 2
		[ r[0:m,j-1], seed ] = i4_sobol ( m, seed )
	return r

def i4_sobol ( dim_num, seed ):

#*****************************************************************************80
#
## I4_SOBOL generates a new quasirandom Sobol vector with each call.
#
#  Discussion:
#
#    The routine adapts the ideas of Antonov and Saleev.
#
#  Licensing:
#
#    This code is distributed under the MIT license.
#
#  Modified:
#
#    22 February 2011
#
#  Author:
#
#    Original FORTRAN77 version by Bennett Fox.
#    MATLAB version by John Burkardt.
#    PYTHON version by Corrado Chisari
#
#  Reference:
#
#    Antonov, Saleev,
#    USSR Computational Mathematics and Mathematical Physics,
#    olume 19, 1980, pages 252 - 256.
#
#    Paul Bratley, Bennett Fox,
#    Algorithm 659:
#    Implementing Sobol's Quasirandom Sequence Generator,
#    ACM Transactions on Mathematical Software,
#    Volume 14, Number 1, pages 88-100, 1988.
#
#    Bennett Fox,
#    Algorithm 647:
#    Implementation and Relative Efficiency of Quasirandom 
#    Sequence Generators,
#    ACM Transactions on Mathematical Software,
#    Volume 12, Number 4, pages 362-376, 1986.
#
#    Ilya Sobol,
#    USSR Computational Mathematics and Mathematical Physics,
#    Volume 16, pages 236-242, 1977.
#
#    Ilya Sobol, Levitan, 
#    The Production of Points Uniformly Distributed in a Multidimensional 
#    Cube (in Russian),
#    Preprint IPM Akad. Nauk SSSR, 
#    Number 40, Moscow 1976.
#
#  Parameters:
#
#    Input, integer DIM_NUM, the number of spatial dimensions.
#    DIM_NUM must satisfy 1 <= DIM_NUM <= 40.
#
#    Input/output, integer SEED, the "seed" for the sequence.
#    This is essentially the index in the sequence of the quasirandom
#    value to be generated.	On output, SEED has been set to the
#    appropriate next value, usually simply SEED+1.
#    If SEED is less than 0 on input, it is treated as though it were 0.
#    An input value of 0 requests the first (0-th) element of the sequence.
#
#    Output, real QUASI(DIM_NUM), the next quasirandom vector.
#
	global atmost
	global dim_max
	global dim_num_save
	global initialized
	global lastq
	global log_max
	global maxcol
	global poly
	global recipd
	global seed_save
	global v

	if ( not 'initialized' in globals().keys() ):
		initialized = 0
		dim_num_save = -1

	if ( not initialized or dim_num != dim_num_save ):
		initialized = 1
		dim_max = 40
		dim_num_save = -1
		log_max = 30
		seed_save = -1
#
#	Initialize (part of) V.
#
		v = zeros((dim_max,log_max))
		v[0:40,0] = transpose([ \
			1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
			1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
			1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
			1, 1, 1, 1, 1, 1, 1, 1, 1, 1 ])

		v[2:40,1] = transpose([ \
			1, 3, 1, 3, 1, 3, 3, 1, \
			3, 1, 3, 1, 3, 1, 1, 3, 1, 3, \
			1, 3, 1, 3, 3, 1, 3, 1, 3, 1, \
			3, 1, 1, 3, 1, 3, 1, 3, 1, 3 ])

		v[3:40,2] = transpose([ \
			7, 5, 1, 3, 3, 7, 5, \
			5, 7, 7, 1, 3, 3, 7, 5, 1, 1, \
			5, 3, 3, 1, 7, 5, 1, 3, 3, 7, \
			5, 1, 1, 5, 7, 7, 5, 1, 3, 3 ])

		v[5:40,3] = transpose([ \
			1, 7, 9,13,11, \
			1, 3, 7, 9, 5,13,13,11, 3,15, \
			5, 3,15, 7, 9,13, 9, 1,11, 7, \
			5,15, 1,15,11, 5, 3, 1, 7, 9 ])
	
		v[7:40,4] = transpose([ \
			9, 3,27, \
			15,29,21,23,19,11,25, 7,13,17, \
			1,25,29, 3,31,11, 5,23,27,19, \
			21, 5, 1,17,13, 7,15, 9,31, 9 ])

		v[13:40,5] = transpose([ \
							37,33, 7, 5,11,39,63, \
		 27,17,15,23,29, 3,21,13,31,25, \
			9,49,33,19,29,11,19,27,15,25 ])

		v[19:40,6] = transpose([ \
			13, \
			33,115, 41, 79, 17, 29,119, 75, 73,105, \
			7, 59, 65, 21,	3,113, 61, 89, 45,107 ])

		v[37:40,7] = transpose([ \
			7, 23, 39 ])
#
#	Set POLY.
#
		poly= [ \
			1,	 3,	 7,	11,	13,	19,	25,	37,	59,	47, \
			61,	55,	41,	67,	97,	91, 109, 103, 115, 131, \
			193, 137, 145, 143, 241, 157, 185, 167, 229, 171, \
			213, 191, 253, 203, 211, 239, 247, 285, 369, 299 ]

		atmost = 2**log_max - 1
#
#	Find the number of bits in ATMOST.
#
		maxcol = i4_bit_hi1 ( atmost )
#
#	Initialize row 1 of V.
#
		v[0,0:maxcol] = 1

#
#	Things to do only if the dimension changed.
#
	if ( dim_num != dim_num_save ):
#
#	Check parameters.
#
		if ( dim_num < 1 or dim_max < dim_num ):
			print('I4_SOBOL - Fatal error!') 
			print('	The spatial dimension DIM_NUM should satisfy:') 
			print('		1 <= DIM_NUM <= %d'%dim_max)
			print('	But this input value is DIM_NUM = %d'%dim_num)
			return

		dim_num_save = dim_num
#
#	Initialize the remaining rows of V.
#
		for i in range(2 , dim_num+1):
#
#	The bits of the integer POLY(I) gives the form of polynomial I.
#
#	Find the degree of polynomial I from binary encoding.
#
			j = poly[i-1]
			m = 0
			while ( 1 ):
				j = math.floor ( j / 2. )
				if ( j <= 0 ):
					break
				m = m + 1
#
#	Expand this bit pattern to separate components of the logical array INCLUD.
#
			j = poly[i-1]
			includ=zeros(m)
			for k in range(m, 0, -1):
				j2 = math.floor ( j / 2. )
				includ[k-1] =  (j != 2 * j2 )
				j = j2
#
#	Calculate the remaining elements of row I as explained
#	in Bratley and Fox, section 2.
#
			for j in range( m+1, maxcol+1 ):
				newv = v[i-1,j-m-1]
				l = 1
				for k in range(1, m+1):
					l = 2 * l
					if ( includ[k-1] ):
						newv = bitwise_xor ( int(newv), int(l * v[i-1,j-k-1]) )
				v[i-1,j-1] = newv
#
#	Multiply columns of V by appropriate power of 2.
#
		l = 1
		for j in range( maxcol-1, 0, -1):
			l = 2 * l
			v[0:dim_num,j-1] = v[0:dim_num,j-1] * l
#
#	RECIPD is 1/(common denominator of the elements in V).
#
		recipd = 1.0 / ( 2 * l )
		lastq=zeros(dim_num)

	seed = int(math.floor ( seed ))

	if ( seed < 0 ):
		seed = 0

	if ( seed == 0 ):
		l = 1
		lastq=zeros(dim_num)

	elif ( seed == seed_save + 1 ):
#
#	Find the position of the right-hand zero in SEED.
#
		l = i4_bit_lo0 ( seed )

	elif ( seed <= seed_save ):

		seed_save = 0
		l = 1
		lastq=zeros(dim_num)

		for seed_temp in range( int(seed_save), int(seed)):
			l = i4_bit_lo0 ( seed_temp )
			for i in range(1 , dim_num+1):
				lastq[i-1] = bitwise_xor ( int(lastq[i-1]), int(v[i-1,l-1]) )

		l = i4_bit_lo0 ( seed )

	elif ( seed_save + 1 < seed ):

		for seed_temp in range( int(seed_save + 1), int(seed) ):
			l = i4_bit_lo0 ( seed_temp )
			for i in range(1, dim_num+1):
				lastq[i-1] = bitwise_xor ( int(lastq[i-1]), int(v[i-1,l-1]) )

		l = i4_bit_lo0 ( seed )
#
#	Check that the user is not calling too many times!
#
	if ( maxcol < l ):
		print('I4_SOBOL - Fatal error!')
		print('	Too many calls!')
		print('	MAXCOL = %d\n'%maxcol)
		print('	L =			%d\n'%l)
		return
#
#	Calculate the new components of QUASI.
#
	quasi=zeros(dim_num)
	for i in range( 1, dim_num+1):
		quasi[i-1] = lastq[i-1] * recipd
		lastq[i-1] = bitwise_xor ( int(lastq[i-1]), int(v[i-1,l-1]) )

	seed_save = seed
	seed = seed + 1

	return [ quasi, seed ]

def i4_uniform ( a, b, seed ):

#*****************************************************************************80
#
## I4_UNIFORM returns a scaled pseudorandom I4.
#
#  Discussion:
#
#    The pseudorandom number will be scaled to be uniformly distributed
#    between A and B.
#
#  Licensing:
#
#    This code is distributed under the MIT license.
#
#  Modified:
#
#    22 February 2011
#
#  Author:
#
#    Original MATLAB version by John Burkardt.
#    PYTHON version by Corrado Chisari
#
#  Reference:
#
#    Paul Bratley, Bennett Fox, Linus Schrage,
#    A Guide to Simulation,
#    Springer Verlag, pages 201-202, 1983.
#
#    Pierre L'Ecuyer,
#    Random Number Generation,
#    in Handbook of Simulation,
#    edited by Jerry Banks,
#    Wiley Interscience, page 95, 1998.
#
#    Bennett Fox,
#    Algorithm 647:
#    Implementation and Relative Efficiency of Quasirandom
#    Sequence Generators,
#    ACM Transactions on Mathematical Software,
#    Volume 12, Number 4, pages 362-376, 1986.
#
#    Peter Lewis, Allen Goodman, James Miller
#    A Pseudo-Random Number Generator for the System/360,
#    IBM Systems Journal,
#    Volume 8, pages 136-143, 1969.
#
#  Parameters:
#
#    Input, integer A, B, the minimum and maximum acceptable values.
#
#    Input, integer SEED, a seed for the random number generator.
#
#    Output, integer C, the randomly chosen integer.
#
#    Output, integer SEED, the updated seed.
#
	if ( seed == 0 ):
		print('I4_UNIFORM - Fatal error!')
		print('	Input SEED = 0!')

	seed = math.floor ( seed )
	a = round ( a )
	b = round ( b )

	seed = mod ( seed, 2147483647 )

	if ( seed < 0 ) :
		seed = seed + 2147483647

	k = math.floor ( seed / 127773 )

	seed = 16807 * ( seed - k * 127773 ) - k * 2836

	if ( seed < 0 ):
		seed = seed + 2147483647

	r = seed * 4.656612875E-10
#
#	Scale R to lie between A-0.5 and B+0.5.
#
	r = ( 1.0 - r ) * ( min ( a, b ) - 0.5 ) + r * ( max ( a, b ) + 0.5 )
#
#	Use rounding to convert R to an integer between A and B.
#
	value = round ( r )

	value = max ( value, min ( a, b ) )
	value = min ( value, max ( a, b ) )

	c = value

	return [ int(c), int(seed) ]

def prime_ge ( n ):

#*****************************************************************************80
#
## PRIME_GE returns the smallest prime greater than or equal to N.
#
#  Example:
#
#      N    PRIME_GE
#
#    -10     2
#      1     2
#      2     2
#      3     3
#      4     5
#      5     5
#      6     7
#      7     7
#      8    11
#      9    11
#     10    11
#
#  Licensing:
#
#    This code is distributed under the MIT license.
#
#  Modified:
#
#    22 February 2011
#
#  Author:
#
#    Original MATLAB version by John Burkardt.
#    PYTHON version by Corrado Chisari
#
#  Parameters:
#
#    Input, integer N, the number to be bounded.
#
#    Output, integer P, the smallest prime number that is greater
#    than or equal to N.	
#
	p = max ( math.ceil ( n ), 2 )
	while ( not isprime ( p ) ):
		p = p + 1

	return p

def isprime(n):

#*****************************************************************************80
#
## IS_PRIME returns True if N is a prime number, False otherwise
#
#  Licensing:
#
#    This code is distributed under the MIT license.
#
#  Modified:
#
#    22 February 2011
#
#  Author:
#
#    Corrado Chisari
#
#  Parameters:
#
#    Input, integer N, the number to be checked.
#
#    Output, boolean value, True or False
#
	if n!=int(n) or n<1:
		return False
	p=2
	while p<n:
		if n%p==0:
			return False
		p+=1

	return True
	



# Copyright (C) 2021 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import unittest
import random


# unit test for regression with some discrete valeus of the state in last dimension
###################################################################################

class testLocalConstDiscrLastDimRegression(unittest.TestCase):

    # One dimensional test
    # Here two possible state for discrete state
    def testRegression1D(self):
        import StOptReg 
        nbSimul = 5000000;
        np.random.seed(1000)
        x = np.zeros(shape=(2,nbSimul))
        # real function
        x[0,:] =  np.random.uniform(-1.,1.,size=(1,nbSimul))
        print(x.dtype)
        stateDiscr =  np.random.randint(2,size=(1,nbSimul))
        x[1,:]  = stateDiscr.astype(float)
        print(x.dtype)
        # real function
        toReal1 = (2+x[0,:]+(1+x[0,:])*(1+x[0,:]))
        toReal2 = (1+x[0,:]+(2+2*x[0,:])*(1+0.1*x[0,:]))
        # function to regress
        toRegress = np.where(stateDiscr[0,:]==0,toReal1, toReal2)
        # mesh (second chosen the number of states  (could be lower))
        nbMesh = np.array([500,2],dtype=np.int32)
        # Regressor
        regressor = StOptReg.LocalConstDiscrLastDimRegression(False,x,nbMesh)
        # nb simul
        nbSimul= regressor.getNbSimul()
        # particles
        part = regressor.getParticles()
        # test particules
        y = regressor.getAllSimulations(toRegress).transpose()
        # compare to real value
        diff = np.max(np.where(stateDiscr[0,:]==0,np.abs(toReal1-y), np.abs(toReal2-y)))
        self.assertAlmostEqual(diff,0., 1,"Difference between function and its condition expectation estimated greater than tolerance")
        # get back basis function
        regressedFuntionCoeff= regressor.getCoordBasisFunction(toRegress)
        # get back all values
        ySecond= regressor.getValues(x,regressedFuntionCoeff).transpose()
        diff = max(abs(y-ySecond))
        self.assertAlmostEqual(diff,0., 7,"Difference between function and its condition expectation estimated greater than tolerance")
                                        
if __name__ == '__main__': 
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np


# Calculate {1 \over { \sqrt{2 \pi}}}  \int_{\infty}^d e^{-x^2/2} dx
class NormalCumulativeDistribution :
    
    # constructor
    def __init__(self) :
        
        return None

    # calculate the value function
    # p_x point where the function is evaluated
    # return function value
    def operator(self, p_x) :
        
        p = 0.2316419
        b1 = 0.319381530
        b2 = -0.356563782
        b3 = 1.781477937
        b4 = -1.821255978
        b5 = 1.330274429

        if p_x > 0. :
            t = 1. / (1. + p * p_x)
            return 1. - np.exp(-p_x * p_x / 2.) * t * (b1 + t * (b2 + t * (b3 + t * (b4 + t * b5)))) / (np.sqrt(2. * np.pi))
        elif p_x < 0. :
            t = 1. / (1. - p * p_x)
            return np.exp(-p_x * p_x / 2.) * t * (b1 + t * (b2 + t * (b3 + t * (b4 + t * b5)))) / (np.sqrt(2. * np.pi))
        else :
            return 0.5



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import unittest
import random
import StOptGrids
import StOptReg
import StOptGeners


# unit test for dumping binary archive of regressed value and Read then
######################################################################

class testBinaryArchiveStOpt(unittest.TestCase):

      
      def testSimpleStorageAndLectureRecGrid(self):
        
        # low value for the mesh
        lowValues =np.array([1.,2.,3.],dtype=np.float64)
        # size of the mesh
        step = np.array([0.7,2.3,1.9],dtype=np.float64)
        # number of step
        nbStep = np.array([4,5,6], dtype=np.int32)
        # degree of the polynomial in each direction
        degree =  np.array([2,1,3], dtype=np.int32)
        # create the Legendre grid
        grid = StOptGrids.RegularLegendreGrid(lowValues,step,nbStep,degree )


        # simulate the perburbed values
        ################################
        nbSimul =40000
        np.random.seed(1000)
        x = np.random.uniform(-1.,1.,size=(1,nbSimul));
        # mesh
        nbMesh = np.array([16],dtype=np.int32)
        # Create the regressor
        #####################
        regressor = StOptReg.LocalLinearRegression(False,x,nbMesh)
        
        # regressed values same values for each point of the grid
        #########################################################
        toReal = (2+x[0,:]+(1+x[0,:])*(1+x[0,:]))
        # function to regress
        toRegress = toReal + 4*np.random.normal(0.,1,nbSimul)
        # create a matrix (number of stock points by number of simulations)
        toRegressMult = np.zeros(shape=(len(toRegress),grid.getNbPoints()))
        for i in range(toRegressMult.shape[1]):
           toRegressMult[:,i] = toRegress        
        # into a list  : two times to test 2 regimes
        listToReg = []
        listToReg.append(toRegressMult)
        listToReg.append(toRegressMult)
    

        # Create the binary archive to dump
        ###################################
        archiveToWrite = StOptGeners.BinaryFileArchive("MyArchive","w")
        # step 1
        archiveToWrite.dumpGridAndRegressedValue("toStore", 1,listToReg, regressor,grid)
        # step 3
        archiveToWrite.dumpGridAndRegressedValue("toStore", 3,listToReg, regressor,grid)


        # Read the regressed values
        ###########################
        archiveToRead =  StOptGeners.BinaryFileArchive("MyArchive","r")
        contValues = archiveToRead.readGridAndRegressedValue(3,"toStore")


        # list of 2 continuation values
        ##############################
        stockPoint = np.array([1.5,3.,5.])
        uncertainty = np.array([0.])
        value =contValues[0].getValue(stockPoint,uncertainty)
      

      # non regular grid
      def testSimpleStorageAndLectureNonRegular(self):
        
        # create the Legendre grid
        grid = StOptGrids.GeneralSpaceGrid([[ 1., 1.7, 2.4, 3.1, 3.8 ],
                                            [2., 4.3, 6.6, 8.9, 11.2, 15.],
                                            [3., 4.9, 5.8, 7.7, 10.,20.]])

        # simulate the perburbed values
        ################################
        nbSimul =40000
        np.random.seed(1000)
        x = np.random.uniform(-1.,1.,size=(1,nbSimul));
        # mesh
        nbMesh = np.array([16],dtype=np.int32)
        # Create the regressor
        #####################
        regressor = StOptReg.LocalLinearRegression(False,x,nbMesh)
        
        # regressed values same values for each point of the grid
        #########################################################
        toReal = (2+x[0,:]+(1+x[0,:])*(1+x[0,:]))
        # function to regress
        toRegress = toReal + 4*np.random.normal(0.,1,nbSimul)
        # create a matrix (number of stock points by number of simulations)
        toRegressMult = np.zeros(shape=(len(toRegress),grid.getNbPoints()))
        for i in range(toRegressMult.shape[1]):
           toRegressMult[:,i] = toRegress        
        # into a list  : two times to test 2 regimes
        listToReg = []
        listToReg.append(toRegressMult)
        listToReg.append(toRegressMult)
    

        # Create the binary archive to dump
        ###################################
        archiveToWrite = StOptGeners.BinaryFileArchive("MyArchive","w")
        # step 1
        archiveToWrite.dumpGridAndRegressedValue("toStore", 1,listToReg, regressor,grid)
        # step 3
        archiveToWrite.dumpGridAndRegressedValue("toStore", 3,listToReg, regressor,grid)


        # Read the regressed values
        ###########################
        archiveToRead =  StOptGeners.BinaryFileArchive("MyArchive","r")
        contValues = archiveToRead.readGridAndRegressedValue(3,"toStore")


        # list of 2 continuation values
        ##############################
        stockPoint = np.array([1.5,3.,5.])
        uncertainty = np.array([0.])
        value =contValues[0].getValue(stockPoint,uncertainty)
        
if __name__ == '__main__': 
    unittest.main()



# Copyright (C) 2016, 2018 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import math as maths
import random
import unittest
import math


# Implement a Black Scholes simulator
class BlackScholesSimulator :
    
    # Constructor for Black Scholes simulator
    # p_initialValues  initial values for assets
    # p_sigma          volatility of assets
    # p_mu             trend for assets
    # p_correl         correlation for assets
    # p_T              maturity
    # p_nbStep         number of time step
    # p_nbSimul        number of simulations
    # p_bForward       true if the simulation is forward, false if backward
    def __init__(self, p_initialValues, p_sigma, p_mu, p_correl, p_T, p_nbStep, p_nbSimul, p_bForward) :
        
        self.m_initialValues = p_initialValues
        self.m_sigma = p_sigma
        self.m_mu = p_mu
        self.corrFactTrans = np.linalg.cholesky(p_correl)
        self.m_T = p_T
        self.m_step = p_T / p_nbStep
        self.m_nbStep = p_nbStep
        self.m_nbSimul = p_nbSimul
        self.m_bForward = p_bForward
        self.m_currentStep = 0. if p_bForward else p_T # Current step during resolution
        self.m_brownian = np.zeros((len(p_initialValues), p_nbSimul)) # store the brownian motion values (no correlation)
        
        np.random.seed(0)
       
        if self.m_bForward :
            self.m_brownian = np.zeros((len(p_initialValues), p_nbSimul))
        else :
            sqrtT = maths.sqrt(self.m_T)
            self.m_brownian = sqrtT * np.random.randn(len(p_initialValues), p_nbSimul)
            
    # gets brownian motions correlated and send back asset values
    def brownianToAsset(self) :
        
        assetToReturn = np.zeros((len(self.m_initialValues), self.m_nbSimul))
        firstTerm = (self.m_mu - 0.5 * self.m_sigma *  self.m_sigma) * self.m_currentStep
        assetToReturn[:,:] = (self.m_initialValues * np.exp(firstTerm + self.m_sigma * self.m_brownian.transpose())).transpose()
         
        return assetToReturn
        
    # a step forward for brownians
    def forwardStepForBrownian(self) :
        
        normalSample = np.random.randn(len(self.m_initialValues), self.m_nbSimul)
        self.m_brownian = (np.multiply(maths.sqrt(self.m_currentStep), normalSample.transpose())).transpose()  # Check shape normalSample =  shape noise

    # a step backward for brownians
    def backwardStepForBrownian(self) :
        
        if abs(self.m_currentStep - 0.) <= (0.01 * abs(0. + self.m_currentStep) * 10) :
            self.m_brownian = np.zeros((len(self.m_initialValues), self.m_nbSimul))
        else :
            util1 = max(self.m_currentStep / (self.m_currentStep + self.m_step), 0.)
            util2 = maths.sqrt(util1 * self.m_step)
            # use brownian bridge
            self.m_brownian = (np.multiply(util1, self.m_brownian) + np.multiply(util2, np.random.randn(len(self.m_initialValues), self.m_nbSimul)))
            
    # get current asset values
    def getParticles(self) :
        
        return self.brownianToAsset()

    # a step forward for simulations
    def stepForward(self) :
        
        if self.m_bForward == False :
            pass
        
        self.m_currentStep += self.m_step
        self.forwardStepForBrownian()

    # return the asset values (asset,simulations)
    def stepBackward(self) :
        
        if self.m_bForward == True :
            pass
        
        self.m_currentStep -= self.m_step
        self.backwardStepForBrownian()

    # a step forward for simulations
    # return the asset values (asset,simulations)
    def stepForwardAndGetParticles(self) :
        
        if self.m_bForward == False :
            pass
        
        self.m_currentStep += self.m_step
        self.forwardStepForBrownian()
        
        return self.brownianToAsset()

    # a step backward for simulations
    # return the asset values (asset,simulations)
    def stepBackwardAndGetParticles(self) :
        
        if self.m_bForward == True :
            pass
        
        self.m_currentStep -= self.m_step
        self.backwardStepForBrownian()
        
        return self.brownianToAsset()

    # Get back attribute
    
    def getCurrentStep(self) :
        
        return self.m_currentStep

    def getT(self) :
       
        return self.m_T

    def getStep(self) :
        
        return self.m_step

    def getInitialValues(self) :
        
        return self.m_initialValues

    def getSigma(self) :
        
        return self.m_sigma

    def getMu(self) :
        
        return self.m_mu

    def getNbSimul(self) :
        
        return self.m_nbSimul

    def getNbStep(self) :
        
        return self.m_nbStep

    def getDimension(self) :
        
        return len(self.m_sigma)

    # actualize at date t=0
    def getActu(self):
        return math.exp(- self.m_mu[0]* self.m_currentStep )

    # actualize on one step
    def getActuStep(self):
        return math.exp(- self.m_mu[0]* self.m_step )
    
class BlackScholesSimulatorTest(unittest.TestCase):
    
    def test_callOption(self):
        
        S_0 = np.zeros(2) + 100.
        vol = np.zeros(2) + 0.25
        mu = np.zeros(2) + 0.1
        corr = np.eye(2)
        T = 2.
        nbStep = 10
        nbSimul = 100000
        K = 100.
        
        ga = BlackScholesSimulator(S_0, vol, mu, corr, T, nbStep, nbSimul, True)
        bu = BlackScholesSimulator(S_0, vol, mu, corr, T, nbStep, nbSimul, False)
        SF = np.zeros(nbStep / 2)
        SB = np.zeros(nbStep / 2)
        
        def compareList(l, res) :
            for i in list(range(len(l))) :
                res[i] = max(l[i], 0.)
            
            return res
     
        for i in range(nbStep / 2) :
            lF = np.zeros(nbSimul)
            compareList(ga.stepForwardAndGetParticles()[0,:] - K,lF)
            SF[i] = np.mean(lF * maths.exp(-mu[0] * T/2))
            lB = np.zeros(nbSimul)
            compareList(bu.stepBackwardAndGetParticles()[0,:] - K,lB)
            SB[i] = np.mean(lB * maths.exp(-mu[0] * T/2))
             
        self.assertAlmostEqual(SF[nbStep / 2 - 1], 15, None, None, 0.2)
        self.assertAlmostEqual(SB[nbStep / 2 - 1], 15, None, None, 0.2)
        self.assertAlmostEqual(SF[nbStep / 2 - 1], SB[nbStep / 2 - 1], None, None, 0.2)
          
    def test_martingale(self):
        
        S_0 = np.zeros(2) + 100.0
        vol = np.zeros(2) + 0.25
        mu = np.zeros(2)
        corr = np.eye(2)
        T = 10.
        nbStep = 10
        nbSimul = 1000000
        K = 100.
        dp = np.zeros(nbStep)
          
        zor = BlackScholesSimulator(S_0, vol, mu, corr, T, nbStep, nbSimul, True)
          
        for j in range(nbStep) :
            dp[j] = np.mean(zor.stepForwardAndGetParticles()[0,:] * maths.exp(-mu[0] * (j + 1)))
            self.assertAlmostEqual(dp[j], S_0[0], None, None, 0.2)
            
if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import unittest
import random
import math
import StOptGrids
import StOptReg


# unit test for continuation values
##################################

class testContValues(unittest.TestCase):

    # unit test for continuation cuts
    def testSimpleGridsAndRegressor(self):
        # low value for the meshes
        lowValues =np.array([1.,2.,3.],dtype=float)
        # size of the meshes
        step = np.array([0.7,2.3,1.9],dtype=float)
        # number of steps
        nbStep = np.array([3,2,4], dtype=np.int32)
        # create the regular grid
        #########################
        grid = StOptGrids.RegularSpaceGrid(lowValues,step,nbStep)
        # simulation
        nbSimul =10000
        np.random.seed(1000)
        x = np.random.uniform(-1.,1.,size=(1,nbSimul));
        # mesh
        nbMesh = np.array([16],dtype=np.int32)
        # Create the regressor
        #####################
        regressor = StOptReg.LocalLinearRegression(False,x,nbMesh)
        # regressed values
        toReal = (2+x[0,:]+(1+x[0,:])*(1+x[0,:]))
        # function to regress
        toRegress = toReal + 4*np.random.normal(0.,1,nbSimul)
        # fictitous cuts with 0 sensibility (1 dimension)
        toRegressCuts = np.zeros(shape=(4*len(toRegress),grid.getNbPoints()))
        for i in range(toRegressCuts.shape[1]):
           toRegressCuts[:len(toRegress),i] = toRegress
        
        # Now create the continuation cut object
        ########################################
        contOb = StOptReg.ContinuationCut(grid,regressor,toRegressCuts)
        hyperCube  = np.array([[lowValues[0],lowValues[0]+step[0]*nbStep[0]],
                               [lowValues[1],lowValues[1]+step[1]*nbStep[1]],
                               [lowValues[2],lowValues[2]+step[2]*nbStep[2]]])
        regressCuts = contOb.getCutsAllSimulations(hyperCube)

                
if __name__ == '__main__': 
    unittest.main()



# Copyright (C) 2018 The Regents of the University of California, Michael Ludkovski and Aditya Maheshwari
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)

from __future__ import division
import numpy as np
import math
import matplotlib.pyplot as plt
import microgridDEA.parameters as bv
import microgridDEA.parameters as bp


def residualDemand(param, nstep, nbsimulOpt, T, X0, typeRMC, sampleType="insample"):


	lambd = param.lambd
	sigma = param.sigma
	A = param.ampl
	mu = param.mu

	dt = T/nstep

	if sampleType=="insample":
		np.random.seed(seed=5)
	elif sampleType=="outsample":
		np.random.seed(seed=96)


	if typeRMC == 'regress now 2D':

		loadMatrix = np.zeros((nbsimulOpt,nstep+1))
		loadMatrix[:,0] = X0

		for i in range(1,nstep+1):

                        
			dW = np.random.normal(0,1,nbsimulOpt)*np.sqrt(dt)

			loadMatrix[:,i] = loadMatrix[:,i-1] + lambd*(mu - loadMatrix[:,i-1])*dt + sigma*dW #sigma*loadMatrix[:,i-1]*dW

		loadMatrix = np.where(loadMatrix>param.DemandUB,param.DemandUB,loadMatrix)


#		inventory = np.random.uniform(param.I_minMax[0],param.I_minMax[1],nbsimulOpt)

		return loadMatrix#, inventory

	if typeRMC == 'gd':


		loadMatrix = np.zeros((nbsimulOpt,nstep+1))
		loadMatrix[:,0] = X0

		for i in range(1,nstep+1):
                        
			dW = np.random.normal(0,1,nbsimulOpt)*np.sqrt(dt)

			loadMatrix[:,i] = loadMatrix[:,i-1] + lambd*(mu - loadMatrix[:,i-1])*dt + sigma*dW #sigma*loadMatrix[:,i-1]*dW

#		inventory = np.linspace(param.I_minMax[0],param.I_minMax[1],param.meshI+1)

		loadMatrix = np.where(loadMatrix>param.DemandUB,param.DemandUB,loadMatrix)

		return loadMatrix#, inventory


if __name__ == '__main__':
    

	param = bp.basicVariables()
	
	nbsimulOpt =5
	X0 = np.linspace(-param.ampl,param.ampl,nbsimulOpt)
	X0[:]=0
	# print len(X0)
	
	demand, inventory = residualDemand(param, param.nstep,  nbsimulOpt, param.maturity, X0, 'gd')
	print(len(inventory))
	# plt.figure(2)
	for simcnt in range(nbsimulOpt):
		plt.plot(demand[simcnt,:])
	plt.grid(True)
	plt.show()





	# param = bv.basicVariables()
	# print param.B_minMax, param.I_minMax[0],param.I_minMax[1], param.d_minMax, param.K

	# X0 = np.arange(-10,10,20/param.nbsimulOpt)
	# # matrix of prices with rows as (number of simulations) and columns as (number of time steps+1)
	# demand, inventory, control = residualDemand(param, param.nstep,  param.nbsimulOpt, param.maturity, X0, 'regress now 2D')

	# for i in range(0,100):
	# 	plt.plot(demand[i,:])
	# plt.show()



# Copyright (C) 2016, 2018 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import math
import random


class AR1Simulator :
    
    # Constructor
    # p_D0     Initial value
    # p_m      average value
    # p_sigma  Volatility
    # p_mr     Mean reverting per factor
    # p_T      Maturity
    # p_nbStep Number of time step for simulation
    # p_nbSimul Number of simulations for the Monte Carlo
    # p_bForward true if the simulator is forward, false if the simulation is backward
    def __init__(self, p_D0, p_m, p_sigma, p_mr, p_T, p_nbStep, p_nbSimul, p_bForward) :
        
        self.m_D0 = p_D0
        self.m_m = p_m
        self.m_sigma = p_sigma
        self.m_mr = p_mr
        self.m_T = p_T
        self.m_step = p_T / p_nbStep
        self.m_nbStep = p_nbStep
        self.m_nbSimul = p_nbSimul
        self.m_bForward = p_bForward
        self.m_currentStep = 0. if p_bForward else p_T
        self.m_OUProcess = np.zeros(p_nbSimul)

        np.random.seed(0)
        
        if self.m_bForward :
            self.m_OUProcess = np.zeros(p_nbSimul)
        
        else :
            stDev = self.m_sigma * math.sqrt(1 - math.exp(-2 * self.m_mr * self.m_T) / (2 * self.m_mr))
            self.m_OUProcess = stDev * np.random.randn(p_nbSimul)
            
    # a step forward for OU process
    def forwardStepForOU(self) :
        
        racine = math.sqrt((1 - np.exp(-2 * self.m_mr * self.m_currentStep)) / (2 * self.m_mr))
        stDev = self.m_sigma * racine
        expActu = np.exp(-self.m_mr * self.m_step)
        normalSample = np.random.randn(self.m_nbSimul)
        increment = np.multiply(stDev, normalSample)
        # update OU process
        self.m_OUProcess = np.multiply(self.m_OUProcess, expActu) + increment
        
    # a step backward for OU process
    def backwardStepForOU(self) :
        
        if self.m_currentStep <= 0. :
            self.m_OUProcess = np.zeros(self.m_nbSimul)
        
        else :
            # use brownian bridge
            util = np.sinh(self.m_mr * self.m_currentStep) / np.sinh(self.m_mr * (self.m_currentStep + self.m_step))
            variance = np.multiply(pow(self.m_sigma, 2.) / (np.multiply(2, self.m_mr)), (1 - np.exp(-2 * self.m_mr * self.m_currentStep)) * pow(1 - np.multiply(np.exp(-self.m_mr * self.m_step), util), 2.) + np.multiply(1 - np.exp(-2 * self.m_mr * self.m_step), pow(util, 2.)))
            stdDev = np.sqrt(variance)
            temp = self.m_OUProcess
            self.m_OUProcess = np.multiply(temp, util) + np.multiply(stdDev, np.random.randn(self.m_nbSimul))

    # get current markov state
    def getParticles(self) :
        
        ret = self.m_OUProcess + (np.zeros(len(self.m_OUProcess)) + ((self.m_D0 - self.m_m) * math.exp(- self.m_m * self.m_currentStep) + self.m_m))
        retMap = np.zeros((1, len(ret)))
        for i in range(len(ret)):
            retMap[0,i] = max(ret[i], 0.)
        
        return retMap
    
    # a step forward for simulations
    def stepForward(self) :
        
        if self.m_bForward == False :
            pass
        
        else :
            self.m_currentStep += self.m_step
            self.forwardStepForOU()

    # return the asset values (asset, simulations)
    def stepBackward(self) :
        
        if self.m_bForward == True :
            pass
        
        else :
            self.m_currentStep -= self.m_step
            self.backwardStepForOU()

    # a step forward for simulations
    # return  the asset values (asset, simulations)
    def stepForwardAndGetParticles(self) :
        
        if self.m_bForward == False :
            pass
        
        else :
            self.m_currentStep += self.m_step
            self.forwardStepForOU()
            
            return self.getParticles()

    # a step backward for simulations
    # return  the asset values (asset, simulations)
    def stepBackwardAndGetParticles(self) :
        
        if self.m_bForward == True :
            pass
        
        else :
            self.m_currentStep -= self.m_step
            self.backwardStepForOU()
            
            return self.getParticles()

    # Get back attribute
    
    def getCurrentStep(self) :
        
        return self.m_currentStep

    def getT(self) :
        
        return self.m_T

    def getStep(self) :
        
        return self.m_step

    def getSigma(self) :
        
        return self.m_sigma

    def getMr(self) :
        
        return self.m_mr

    def getNbSimul(self) :
        
        return self.m_nbSimul

    def getNbStep(self) :
        
        return self.m_nbStep

    def getDimension(self) :
        
        return 1



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import math
import utils.OneDimRegularSpaceGrid as rsg
import utils.OneDimData as data
import simulators.MeanRevertingSimulator as mrsim
import dp.OptimizeGasStorageSwitchingCost as ogssc
import StOptReg as reg
import StOptGrids
import dp.DynamicProgrammingByRegression as dyn
import unittest
import importlib

accuracyClose = 0.5


class ZeroFunction:
    
    def __init__(self):
        
        return None
    
    def set(self, a, b, c):
        
        return 0.
    
def testGasStorage() :
        
    # test MPI
    moduleMpi4Py=importlib.util.find_spec('mpi4py')
    if (moduleMpi4Py is not None):
       from mpi4py import MPI
       #########
       maxLevelStorage = 360000
       injectionRateStorage = 60000
       withdrawalRateStorage = 45000
       injectionCostStorage = 0.35
       withdrawalCostStorage = 0.35
       switchingCostStorage =  4.

       maturity = 1.
       nstep = 10

       # define a a time grid
       timeGrid = rsg.OneDimRegularSpaceGrid(np.zeros(1), maturity / nstep, nstep)
       futValues = np.zeros(nstep + 1)
       # periodicity factor
       iPeriod = 52

       for i in range(nstep + 1):
           futValues[i] = 50. + 20 * math.sin((math.pi * i * iPeriod) / nstep)

       # define the future curve
       futureGrid = data.OneDimData(timeGrid, futValues)
       # one dimensional factors
       nDim = 1
       sigma = np.zeros(nDim) + 0.94
       mr = np.zeros(nDim) + 0.29
       # number of simulations
       nbsimulOpt = 20000
       # grid
       #####
       nGrid = 40
       lowValues = np.zeros(1, dtype = float)
       step = np.zeros(1, dtype = float) + maxLevelStorage / nGrid
       nbStep = np.zeros(1, dtype = np.int32) + nGrid
       grid = StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)

       # no actualization
       rate=0.
       # a backward simulator
       ######################
       bForward = False
       backSimulator = mrsim.MeanRevertingSimulator(futureGrid, sigma, mr, rate,maturity, nstep, nbsimulOpt, bForward)
       # optimizer
       ############
       tval = np.zeros(2)
       tval[0] = 0.
       tval[1] = 1.e30
       timeGrid = StOptGrids.OneDimSpaceGrid(tval)
       values = np.zeros(2, dtype = np.int32)
       values[0] = 3
       values[1] = 3
       regime = data.OneDimData(timeGrid, values)
       storage = ogssc.OptimizeGasStorageSwitchingCost(injectionRateStorage, withdrawalRateStorage, injectionCostStorage, withdrawalCostStorage, switchingCostStorage, regime)
       # regressor
       ##########
       nMesh = 4
       nbMesh = np.zeros(1, dtype = np.int32) + nMesh
       regressor = reg.LocalLinearRegression(nbMesh)
       # final value
       vFunction = ZeroFunction()

       # initial values
       initialStock = np.zeros(1) + maxLevelStorage
       initialRegime = 0 # here do nothing (no injection, no withdrawal)

       # Optimize
       ###########
       fileToDump = "CondExpGasSwiCost"
       # link the simulations to the optimizer
       storage.setSimulator(backSimulator)
       valueOptim = dyn.DynamicProgrammingByRegression(grid, storage, regressor, vFunction, initialStock, initialRegime, fileToDump)
       print("valOP", valueOptim)

class testGasStorageSwitchCostTest(unittest.TestCase):
         
    def test_gasStorageSwitchCost(self):
        
        testGasStorage()
    
if __name__ == '__main__':
    
    unittest.main()



# Copyright (C) 2019 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import StOptTree 
import StOptGrids
import StOptGeners
import StOptGlobal


# Simulate the optimal strategy , threaded version, for tree
# p_grid                   grid used for  deterministic state (stocks for example)
# p_optimize               optimizer defining the optimization between two time steps
# p_funcFinalValue         function defining the final value
# p_pointStock             initial point stock
# p_initialRegime          regime at initial date
# p_fileToDump             name of the file used to dump continuation values in optimization
def SimulateTreeControlHighLevel(p_grid, p_optimize, p_funcFinalValue, p_pointStock, p_initialRegime, p_fileToDump) :

    simulator = p_optimize.getSimulator()
    nbStep = simulator.getNbStep()
    states = []
    for i in range(simulator.getNbSimul()) :
        states.append(StOptGlobal.StateTreeStocks(p_initialRegime, p_pointStock, 0))
    ar = StOptGeners.BinaryFileArchive(p_fileToDump, "r")
    # name for continuation object in archive
    nameAr = "Continuation"
    # cost function
    costFunction = np.zeros((p_optimize.getSimuFuncSize(), simulator.getNbSimul()))
    
    # iterate on time steps
    for istep in range(nbStep) :
        NewState = StOptGlobal.SimulateStepTreeControl(ar, nbStep - 1 - istep, nameAr, p_grid, p_optimize).oneStep(states, costFunction)
        # different from C++
        states = NewState[0]
        costFunction = NewState[1]
     
        # new stochastic state
        simulator.stepForward()
        
        for i in range(simulator.getNbSimul()) :
            states[i].setStochasticRealization(simulator.getNodeAssociatedToSim(i))
    # final : accept to exercise if not already done entirely
    for i in range(simulator.getNbSimul()) :
        costFunction[0,i] += p_funcFinalValue.set(states[i].getRegime(), states[i].getPtStock(),simulator.getValueAssociatedToNode(states[i].getStochasticRealization()))
        
    # average gain/cost
    return costFunction.mean()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import unittest
import random

# unit test for regression 
#########################

class testLocalLinearRegression(unittest.TestCase):

    # One dimensional test
    def testRegression1D(self):
        import StOptReg 
        nbSimul = 5000000;
        np.random.seed(1000)
        x = np.random.uniform(-1.,1.,size=(1,nbSimul));
        # real function
        toReal = (2+x[0,:]+(1+x[0,:])*(1+x[0,:]))
        # function to regress
        toRegress = toReal + 4*np.random.normal(0.,1,nbSimul)
        # mesh
        nbMesh = np.array([16],dtype=np.int32)
        # Regressor
        regressor = StOptReg.LocalLinearRegression(False,x,nbMesh)
        # nb simul
        nbSimul= regressor.getNbSimul()
        # particles
        part = regressor.getParticles()
        # test particules
        y = regressor.getAllSimulations(toRegress).transpose()
        # compare to real value
        diff = max(abs(y-toReal))
        self.assertAlmostEqual(diff,0., 1,"Difference between function and its condition expectation estimated greater than tolerance")
        # get back basis function
        regressedFuntionCoeff= regressor.getCoordBasisFunction(toRegress)
        # get back all values
        ySecond= regressor.getValues(x,regressedFuntionCoeff).transpose()
        diff = max(abs(y-ySecond))
        self.assertAlmostEqual(diff,0., 7,"Difference between function and its condition expectation estimated greater than tolerance")


    # two dimensonnal test
    def testRegression2D(self):
        import StOptReg 
        np.random.seed(1000)
        nbSimul = 10000000;
        x = np.random.uniform(-1.,1.,size=(2,nbSimul));
        # real function
        toReal = (2+x[0,:]+(1+x[0,:])*(1+x[0,:]))*(2+x[1,:]+(1+x[1,:])*(1+x[1,:]))
        # function to regress
        toRegress = toReal + 4*np.random.normal(0.,1,nbSimul)
        # mesh
        nbMesh = np.array([16,16],dtype=np.int32)
        # Regressor
        regressor = StOptReg.LocalLinearRegression(False,x,nbMesh)
        y = regressor.getAllSimulations(toRegress).transpose()
        # particles
        part = regressor.getParticles()
        # compare to real value
        diff = max(abs(y-toReal))
        self.assertAlmostEqual(diff,0., 0,"Difference between function and its condition expectation estimated greater than tolerance")
        # get back basis function
        regressedFuntionCoeff= regressor.getCoordBasisFunction(toRegress)
        # get back all values
        ySecond= regressor.getValues(x,regressedFuntionCoeff).transpose()
        diff = max(abs(y-ySecond))
        self.assertAlmostEqual(diff,0., 7,"Difference between function and its condition expectation estimated greater than tolerance")

    # test different second member
    def testSecondMember(self):
         import StOptReg 
         np.random.seed(1000)
         nbSimul = 100;
         x = np.random.uniform(-1.,1.,size=(1,nbSimul));
         toReal = np.array([((2+x[0,:]+(1+x[0,:])*(1+x[0,:]))).tolist(),((3+x[0,:]+(2+x[0,:])*(2+x[0,:]))).tolist()])
         # function to regress
         toRegress = np.array([(toReal[0,:] + 4*np.random.normal(0.,1,nbSimul)).tolist(),(toReal[1,:] + 4*np.random.normal(0.,1,nbSimul)).tolist()])
         # mesh
         nbMesh = np.array([4],dtype=np.int32)
         # Regressor
         regressor = StOptReg.LocalLinearRegression(False,x,nbMesh)
         y = regressor.getAllSimulationsMultiple(toRegress)
         # check 
         z1 = regressor.getAllSimulations(toRegress[0,:]).transpose()
         z2 = regressor.getAllSimulations(toRegress[1,:]).transpose()
         # compare to real value
         diff = max(abs(y[0,:]-z1[:] ))+max(abs(y[1,:]-z2[:] ))
         self.assertAlmostEqual(diff,0., 7,"Difference between function and its condition expectation estimated greater than tolerance")


    # test  date 0
    def testDateZero(self):
        import StOptReg 
        nbSimul = 50000;
        np.random.seed(1000)
        x = np.random.uniform(-1.,1.,size=(1,nbSimul));
        # real function
        toReal = (2+x[0,:]+(1+x[0,:])*(1+x[0,:]))
        # function to regress
        toRegress = toReal + 4*np.random.normal(0.,1,nbSimul)
        # mesh
        nbMesh = np.array([16],dtype=np.int32)
        # Regressor
        regressor = StOptReg.LocalLinearRegression(True,x,nbMesh)
        y = regressor.getAllSimulations(toRegress).mean()  
        self.assertAlmostEqual(y,toRegress.mean(),8,"Not equality by average")
        # get back basis function
        regressedFuntionCoeff= regressor.getCoordBasisFunction(toRegress)
        # get back all values
        ySecond= regressor.getValues(x,regressedFuntionCoeff).transpose()
        diff = max(abs(y-ySecond))
        self.assertAlmostEqual(diff,0., 7,"Difference between function and its condition expectation estimated greater than tolerance")

if __name__ == '__main__': 
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import StOptGrids 
import StOptReg
import numpy as np



# Defines a simple gas storage for optimization and simulation
# No constraint on the storage at the end of optimization period (so the storage will be empty)
class OptimizeGasStorage :
    
    # Constructor
    # p_maxLevel          size of the storage
    # p_injectionRate     injection rate per time step
    # p_withdrawalRate    withdrawal rate between two time steps
    # p_injectionCost     injection cost
    # p_withdrawalCost    withdrawal cost
    def __init__(self, p_injectionRate, p_withdrawalRate, p_injectionCost, p_withdrawalCost) :
        
        self.m_injectionRate = p_injectionRate
        self.m_withdrawalRate = p_withdrawalRate
        self.m_injectionCost = p_injectionCost
        self.m_withdrawalCost = p_withdrawalCost
        
    # define the diffusion cone for parallelism
    # p_regionByProcessor         region (min max) treated by the processor for the different regimes treated
    # returns in each dimension the min max values in the stock that can be reached from the grid p_gridByProcessor for each regime
    def getCone(self, p_regionByProcessor) :
        
        extrGrid = np.zeros(1)
        extrGrid[0] = p_regionByProcessor[0] - self.m_withdrawalRate
        extrGrid[1] = p_regionByProcessor[1] + self.m_injectionRate
        
        return extrGrid
    
        
    # defines a step in optimization
    # p_grid      grid at arrival step after command
    # p_stock     coordinate of the stock point to treat
    # p_condEsp   conditional expectation operator
    # p_phiIn     for each regime  gives the solution calculated at the previous step ( next time step by Dynamic Programming resolution)
    # for each regimes (column) gives the solution for each particle (row)
    def stepOptimize(self, p_grid, p_stock, p_condEsp, p_phiIn) :
        
        nbSimul = self.m_simulator.getNbSimul()
        actuStep = self.m_simulator.getActuStep()
        solutionAndControl = list(range(2))
        solutionAndControl[0] = np.zeros((nbSimul, 1))
        solutionAndControl[1] = np.zeros((nbSimul, 1))
        # Spot price
        spotPrice = self.m_simulator.fromParticlesToSpot(self.m_simulator.getParticles())
        # level if injection
        # size of the stock
        maxStorage = p_grid.getExtremeValues()[0][1]
        injectionMax = min(maxStorage - p_stock[0], self.m_injectionRate)
        injectionStock = p_stock + injectionMax
        # level if withdrawal
        # level min of the stock
        minStorage = p_grid.getExtremeValues()[0][0]
        withdrawalMax = min(p_stock[0] - minStorage, self.m_withdrawalRate)
        withdrawalStock = p_stock - withdrawalMax
        
        if (injectionMax < -1e-10):
            solutionAndControl[0].fill(-1.e30)
            solutionAndControl[1].fill(0.)
            
            # not an admissible point
            return solutionAndControl
        
        condExpSameStock = []
        
        # Suppose that non injection and no withdrawal
        #############################################
        # create interpolator at current stock point
        if p_grid.isInside(p_stock):
            interpolatorCurrentStock = p_grid.createInterpolator(p_stock)
            # cash flow at current stock and previous step
            cashSameStock = interpolatorCurrentStock.applyVec(p_phiIn[0])
            # conditional expectation at current stock point
            condExpSameStock = actuStep * p_condEsp[0].getAllSimulations(interpolatorCurrentStock).squeeze()
            
        # injection
        ###########
        gainInjection = []
        
        if (injectionMax > 0):
            # interpolator for stock level if injection
            interpolatorInjectionStock = p_grid.createInterpolator(injectionStock)
            # cash flow  at previous step at injection level
            cashInjectionStock = interpolatorInjectionStock.applyVec(p_phiIn[0]).squeeze()
            # conditional expectation at injection stock level
            condExpInjectionStock = actuStep * p_condEsp[0].getAllSimulations(interpolatorInjectionStock).squeeze()
            # instantaneous gain if injection
            gainInjection = -injectionMax * (spotPrice + self.m_injectionCost)
        
        # withdrawal
        ############
        gainWithdrawal = []
        
        if (withdrawalMax >0):
            # interpolator for stock level if withdrawal
            interpolatorWithdrawalStock = p_grid.createInterpolator(withdrawalStock)
            # cash flow  at previous step at injection level
            cashWithdrawalStock = interpolatorWithdrawalStock.applyVec(p_phiIn[0]).squeeze()
            # conditional expectation at withdrawal stock level
            condExpWithdrawalStock = actuStep * p_condEsp[0].getAllSimulations(interpolatorWithdrawalStock).squeeze()
            # instantaneous gain if withdrawal
            gainWithdrawal = withdrawalMax * (spotPrice - self.m_withdrawalCost)
        
        # do the arbitrage
        ##################
        if (len(gainWithdrawal) > 0) & (len(gainInjection) > 0):
            # all point admissible
            solutionAndControl[0][:,0] = np.multiply(actuStep, cashSameStock).transpose()
            solutionAndControl[1][:,0] = 0.
            espCondInjection = gainInjection + condExpInjectionStock
            espCondInjectionSuperiorToCondExpSameStock = espCondInjection > condExpSameStock
            
            solutionAndControl[0][:,0] = np.where(espCondInjectionSuperiorToCondExpSameStock, gainInjection + actuStep * cashInjectionStock, solutionAndControl[0][:,0])
            solutionAndControl[1][:,0] = np.where(espCondInjectionSuperiorToCondExpSameStock, injectionMax,  solutionAndControl[1][:,0])
            condExpSameStock = np.where(espCondInjectionSuperiorToCondExpSameStock, espCondInjection, condExpSameStock)
            
            espCondWithdrawal = gainWithdrawal + condExpWithdrawalStock
            
            solutionAndControl[0][:,0] = np.where(espCondWithdrawal > condExpSameStock, gainWithdrawal + actuStep * cashWithdrawalStock, solutionAndControl[0][:,0])
            solutionAndControl[1][:,0] = np.where(espCondWithdrawal > condExpSameStock, - withdrawalMax,  solutionAndControl[1][:,0])
            
        elif len(gainWithdrawal) > 0:
            
            if len(condExpSameStock) > 0:
                solutionAndControl[0][:,0] = np.multiply(actuStep, cashSameStock).transpose()
                solutionAndControl[1][:,0] = 0.
                espCondWithdrawal = gainWithdrawal + condExpWithdrawalStock
                
                solutionAndControl[0][:,0] = np.where(espCondWithdrawal > condExpSameStock, gainWithdrawal + actuStep * cashWithdrawalStock, solutionAndControl[0][:,0])
                solutionAndControl[1][:,0] = np.where(espCondWithdrawal > condExpSameStock, - withdrawalMax, solutionAndControl[1][:,0])
                    
            else:
                solutionAndControl[0][:,0] = gainWithdrawal + actuStep * cashWithdrawalStock
                solutionAndControl[1][:,0] = - withdrawalMax
                
        elif len(gainInjection) > 0:
            
            if len(condExpSameStock) > 0:
                solutionAndControl[0][:,0] = np.multiply(actuStep, cashSameStock).transpose()
                solutionAndControl[1][:,0] = 0.
                espCondInjection = gainInjection + condExpInjectionStock
                espCondInjectionSuperiorToCondExpSameStock = espCondInjection > condExpSameStock
                
                solutionAndControl[0][:,0] = np.where(espCondInjectionSuperiorToCondExpSameStock, gainInjection + actuStep * cashInjectionStock, solutionAndControl[0][:,0])
                solutionAndControl[1][:,0] = np.where(espCondInjectionSuperiorToCondExpSameStock, injectionMax, solutionAndControl[1][:,0])
                
            else:
                solutionAndControl[0][:,0] = gainInjection + actuStep * cashInjectionStock
                solutionAndControl[1][:,0] = injectionMax
                
        return solutionAndControl
        
    # get number of regimes
    def getNbRegime(self) :
  
        return 1
    
    # number of controls
    def getNbControl(self):
        
        return 1
    
    # defines a step in simulation
    # Notice that this implementation is not optimal. In fact no interpolation is necessary for this asset.
    # This implementation is for test and example purpose
    # p_grid          grid at arrival step after command
    # p_continuation  defines the continuation operator for each regime
    # p_state         defines the state value (modified)
    # p_phiInOut      defines the value function (modified)
    def stepSimulate(self, p_grid, p_continuation, p_state, p_phiInOut) :
        actuStep = self.m_simulator.getActuStep()
        actu = self.m_simulator.getActu()
        # optimal stock attained
        ptStockCur = p_state.getPtStock()
        control = np.zeros_like(ptStockCur)
        # spot price
        spotPrice = self.m_simulator.fromOneParticleToSpot(p_state.getStochasticRealization())
        espCondMax = -1.e30
        
        if p_grid.isInside(ptStockCur):
            continuationDoNothing = actuStep * p_continuation[0].getValue(p_state.getPtStock(), p_state.getStochasticRealization())
            espCondMax = continuationDoNothing
            
        # gain to add at current point
        phiAdd = 0
        # size of the stock
        maxStorage = p_grid.getExtremeValues()[0][1]
        # if injection
        injectionMax = min(maxStorage - p_state.getPtStock()[0], self.m_injectionRate)
        
        if injectionMax>0:
            continuationInjection = actuStep * p_continuation[0].getValue(p_state.getPtStock() + injectionMax, p_state.getStochasticRealization())
            gainInjection = - injectionMax * (spotPrice + self.m_injectionCost)
            espCondInjection = gainInjection + continuationInjection
            
            if espCondInjection > espCondMax :
                espCondMax = espCondInjection
                phiAdd = gainInjection
                control[0] = injectionMax
                
        # if withdrawal
        # level min of the stock
        minStorage = p_grid.getExtremeValues()[0][0]
        withdrawalMax = min(p_state.getPtStock()[0] - minStorage, self.m_withdrawalRate)
        
        if withdrawalMax >0:
            gainWithdrawal = withdrawalMax * (spotPrice - self.m_withdrawalCost)
            continuationWithdrawal = actuStep * p_continuation[0].getValue(p_state.getPtStock() - withdrawalMax, p_state.getStochasticRealization())
            espCondWithdrawal = gainWithdrawal + continuationWithdrawal
            if espCondWithdrawal > espCondMax :
                espCondMax = espCondWithdrawal
                phiAdd = gainWithdrawal
                control[0] = -withdrawalMax
        # for return
        p_state.setPtStock(ptStockCur+control)
        p_phiInOut += phiAdd * actu
         
    # Defines a step in simulation using interpolation in controls
    # p_grid          grid at arrival step after command
    # p_control       defines the controls
    # p_state         defines the state value (modified)
    # p_phiInOut      defines the value function (modified): size number of functions to follow
    def stepSimulateControl(self, p_grid, p_control, p_state, p_phiInOut):
        actu = self.m_simulator.getActu()
        ptStock = p_state.getPtStock()
        # spot price
        spotPrice = self.m_simulator.fromOneParticleToSpot(p_state.getStochasticRealization())
        # optimal control
        control = p_control[0].getValue(p_state.getPtStock(), p_state.getStochasticRealization())
        maxStorage = p_grid.getExtremeValues()[0][1]
        minStorage = p_grid.getExtremeValues()[0][0]
        control = max(min(maxStorage - ptStock[0], control), minStorage - ptStock[0])
        
        if control > 0:
            p_phiInOut[0] -= control * (spotPrice + self.m_injectionCost) * actu
        
        elif control < 0:
            p_phiInOut[0] -= control * (spotPrice - self.m_withdrawalCost) * actu
            
        ptStock[0] += control
        p_state.setPtStock(ptStock)
        
    # get size of the  function to follow in simulation
    def getSimuFuncSize(self):
        
        return 1
        
    # store the simulator
    def setSimulator(self, p_simulator):
        
        self.m_simulator = p_simulator
        
    # get the simulator back
    def getSimulator(self):
        
        return self.m_simulator



# Copyright (C) 2021 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import StOptGeners
import StOptGlobal
import imp


# Simulate the optimal strategy with switching
# p_grids                  grid used for  int deterministic state for each regime
# p_optimize               optimizer defining the optimization between two time steps
# p_funcFinalValue         function defining the final value
# p_pointStock             initial point stock
# p_initialRegime          regime at initial date
# p_fileToDump             name of the file used to dump continuation values in optimization
def SimulateSwitchingRegression(p_grids, p_optimize,  p_pointState, p_initialRegime, p_fileToDump) :
    
    simulator = p_optimize.getSimulator()
    nbStep = simulator.getNbStep()
    states = []
    particle0 =  simulator.getParticles()[:,0]
       
    for i in range(simulator.getNbSimul()) :
        states.append(StOptGlobal.StateWithIntState(p_initialRegime, p_pointState, particle0))
            
            
    ar = StOptGeners.BinaryFileArchive(p_fileToDump, "r")
    # name for continuation object in archive
    nameAr = "ContinuationSwitching"
    # cost function
    costFunction = np.zeros((p_optimize.getSimuFuncSize(), simulator.getNbSimul()))

    # iterate on time steps
    for istep in range(nbStep) :
        NewState = StOptGlobal.SimulateStepSwitch(ar, istep, nameAr, p_grids, p_optimize).oneStep(states, costFunction)
        # different from C++
        states = NewState[0]
        costFunction = NewState[1]
        # new stochastic state
        particules = simulator.stepForwardAndGetParticles()
        # update stochastic state
        for i in range(simulator.getNbSimul()) :
            states[i].setStochasticRealization(particules[:,i])


    # average gain/cost
    return costFunction.mean()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import StOptGrids
import StOptSDDP
import StOptGlobal
import Utils
import SDDPSimulators as sim
import SDDPOptimizers as opt
import numpy as NP
import unittest
import math
import imp
import sddp.backwardForwardSDDP as bfSDDP # import of the function written in python

# unitest equivalent of testDemandSDDP : here MPI version
# High level python interface : at level of the backwardForwardSDDP c++ file
############################################################################
def demandSDDPFunc(p_sigD, p_sampleOptim ,p_sampleCheckSimul):

        maturity = 40
        nstep = 40;
        
        # optimizer parameters
        kappaD = 0.2; # mean reverting coef of demand
        spot = 3 ; # spot price

         # define a a time grid
        timeGrid = StOptGrids.OneDimRegularSpaceGrid(0., maturity / nstep, nstep)
        
        # periodicity factor
        iPeriod = 52;
        # average demande values
        demValues = []

        for i in list(range(nstep + 1)) :
            demValues.append(2. + 0.4 * math.cos((math.pi * i * iPeriod) / nstep))
 
        # define average demand
        demGrid = Utils.FutureCurve(timeGrid, demValues)
 
        initialState = demGrid.get(0.)*NP.ones(1)

        finCut = StOptSDDP.SDDPFinalCut(NP.zeros((2,1)))

        # here cuts are not conditional to an uncertainty
        nbMesh = NP.array([],NP.int32)
        nbUncertainties = 1;
        
        # backward simulator
        backwardSimulator = sim.SimulatorGaussianSDDP(nbUncertainties,p_sampleOptim)
        # forward simulator
        forwardSimulator = sim.SimulatorGaussianSDDP(nbUncertainties)
         
        # Create the optimizer
        optimizer = opt.OptimizeDemandSDDP(p_sigD, kappaD,  demGrid, spot, backwardSimulator, forwardSimulator)

        # optimisation dates
        dates = NP.linspace( 0., maturity,nstep + 1);

        # names for archive
        nameRegressor = "RegressorDemand";
        nameCut = "CutDemand";
        nameVisitedStates = "VisitedStateDemand";

        #  precision parameter
        nIterMax = 40
        accuracyClose =  1.
        accuracy = accuracyClose / 100.
        nstepIterations = 4; # check for convergence between nstepIterations step

        values = StOptSDDP.backwardForwardSDDP(optimizer,  p_sampleCheckSimul, initialState, finCut, dates,  nbMesh, nameRegressor, nameCut, nameVisitedStates, nIterMax,
                                                    accuracy, nstepIterations);

        print("Values " , values)
        return values


# unitest equivalent of testDemandSDDP : here low interface python version
# Low level python interface : use backwardForwardSDDP.py
##########################################################################
def demandSDDPFuncLowLevel(p_sigD, p_sampleOptim ,p_sampleCheckSimul):

        maturity = 40
        nstep = 40;
        
        # optimizer parameters
        kappaD = 0.2; # mean reverting coef of demand
        spot = 3 ; # spot price

         # define a a time grid
        timeGrid = StOptGrids.OneDimRegularSpaceGrid(0., maturity / nstep, nstep)
        
 
        # periodicity factor
        iPeriod = 52;
        # average demande values
        demValues = []

        for i in list(range(nstep + 1)) :
            demValues.append(2. + 0.4 * math.cos((math.pi * i * iPeriod) / nstep))
 
        # define average demand
        demGrid =Utils.FutureCurve(timeGrid, demValues)
 
        initialState = demGrid.get(0.)*NP.ones(1)

        finCut = StOptSDDP.SDDPFinalCut(NP.zeros((2,1)))

        # here cuts are not conditional to an uncertainty
        nbMesh = NP.array([],NP.int32)
        nbUncertainties = 1;
        
        # backward simulator
        backwardSimulator = sim.SimulatorGaussianSDDP(nbUncertainties,p_sampleOptim)
        # forward simulator
        forwardSimulator = sim.SimulatorGaussianSDDP(nbUncertainties)
         
        # Create the optimizer
        optimizer = opt.OptimizeDemandSDDP(p_sigD, kappaD,  demGrid, spot, backwardSimulator, forwardSimulator)

        # optimisation dates
        dates = NP.linspace( 0., maturity,nstep + 1);

        # names for archive
        nameRegressor = "RegressorDemand";
        nameCut = "CutDemand";
        nameVisitedStates = "VisitedStateDemand";

        #  precision parameter
        nIterMax = 40
        accuracyClose =  1.
        accuracy = accuracyClose / 100.
        nstepIterations = 4; # check for convergence between nstepIterations step

        values = bfSDDP.backwardForwardSDDP(optimizer,  p_sampleCheckSimul, initialState, finCut, dates,  nbMesh, nameRegressor,
                                     nameCut, nameVisitedStates, nIterMax,
                                     accuracy, nstepIterations);

        return values
  

class testDemandSDDP(unittest.TestCase):
    def testDemandSDDP1D(self):
        try:
            imp.find_module('mpi4py')
            found = True
        except:
            print("Not parallel module found ")
            found = False
    
        if found :
            from mpi4py import MPI
            world = MPI.COMM_WORLD

            sigD = 0.6 ;
            sampleOptim = 500;
            sampleCheckSimul = 500;
            
            values = demandSDDPFunc(sigD, sampleOptim ,sampleCheckSimul)

            if (world.rank==0):
                    print("Values is ",values)
        
    def testDemandSDDP1DLowLevel(self):
        sigD = 0.6 ;
        sampleOptim = 500;
        sampleCheckSimul = 500;
        demandSDDPFuncLowLevel(sigD, sampleOptim ,sampleCheckSimul)


if __name__ == '__main__': 
    unittest.main() 



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import unittest
import random
import math
import StOptGrids
 
# unit test for regular grids
############################

class testGrids(unittest.TestCase):

    # 3 dimensional test for linear interpolation on regular grids
    def testRegularGrids(self):
        # low value for the meshes
        lowValues =np.array([1.,2.,3.],dtype=float)
        # size of the meshes
        step = np.array([0.7,2.3,1.9],dtype=float)
        # number of steps
        nbStep = np.array([4,5,6], dtype=np.int32)
        # create the regular grid
        grid = StOptGrids.RegularSpaceGrid(lowValues,step,nbStep)
        iterGrid = grid.getGridIterator()
        # array to store
        data = np.empty(grid.getNbPoints())
        # iterates on points and store values
        while( iterGrid.isValid()):
          #get coordinates of the point
          pointCoord = iterGrid.getCoordinate()
          data[iterGrid.getCount()] = math.log(1. + pointCoord.sum())
          iterGrid.next()
        # get back an interpolator
        ptInterp = np.array([2.3,3.2,5.9],dtype=float)
        interpol = grid.createInterpolator(ptInterp)
        # calculate interpolated value
        interpValue = interpol.apply(data)
        print(("Interpolated value" , interpValue))
        # test grids function
        iDim = grid.getDimension()
        pt = grid.getExtremeValues()

if __name__ == '__main__': 
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import StOptGeners
import StOptGlobal
import imp


# Simulate the optimal strategy , threaded version
# p_grid                   grid used for  deterministic state (stocks for example)
# p_optimize               optimizer defining the optimization between two time steps
# p_funcFinalValue         function defining the final value
# p_pointStock             initial point stock
# p_initialRegime          regime at initial date
# p_fileToDump             name of the file used to dump continuation values in optimization
def SimulateRegressionDist(p_grid, p_optimize, p_funcFinalValue, p_pointStock, p_initialRegime, p_fileToDump, p_bOneFile) :
    
    try:
        imp.find_module('mpi4py')
        found = True
    except:
        print("Not parallel module found ")
        found = False
        
    if found :
        from mpi4py import MPI
        # from the optimizer get back the simulation
        world = MPI.COMM_WORLD
        simulator = p_optimize.getSimulator()
        nbStep = simulator.getNbStep()
        states = []
        particle0 =  simulator.getParticles()[:,0]
       
        for i in range(simulator.getNbSimul()) :
            states.append(StOptGlobal.StateWithStocks(p_initialRegime, p_pointStock, particle0))
            
        toDump = p_fileToDump
        
        if p_bOneFile is False:
            toDump += "_" + str(world.rank)
            
        ar = StOptGeners.BinaryFileArchive(toDump, "r")
        # name for continuation object in archive
        nameAr = "Continuation"
        # cost function
        costFunction = np.zeros((p_optimize.getSimuFuncSize(), simulator.getNbSimul()))
        
        # iterate on time steps
        for istep in range(nbStep) :
            NewState = StOptGlobal.SimulateStepRegressionDist(ar, istep, nameAr, p_grid, p_optimize, p_bOneFile).oneStep(states, costFunction)
            # different from C++
            states = NewState[0]
            costFunction = NewState[1]
            # new stochastic state
            particules = simulator.stepForwardAndGetParticles()
            
            for i in range(simulator.getNbSimul()) :
                states[i].setStochasticRealization(particules[:,i])
             
        # final : accept to exercise if not already done entirely
        for i in range(simulator.getNbSimul()) :
            costFunction[0,i] += p_funcFinalValue.set(states[i].getRegime(), states[i].getPtStock(), states[i].getStochasticRealization()) * simulator.getActu()
            
        # average gain/cost
        return costFunction.mean()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np


# Defines a simple gas storage for optimization and simulation
# No constraint on the storage at the end of optimization period (so the storage will be empty)
class OptimizeLake :

    # Constructor
    # p_withdrawalRate    withdrawal rate between two time steps
    def __init__(self, p_withdrawalRate) :
        self.m_withdrawalRate = p_withdrawalRate

    # define the diffusion cone for parallelism
    # p_regionByProcessor         region (min max) treated by the processor for the different regimes treated
    # returns in each dimension the min max values in the stock that can be reached from the grid p_gridByProcessor for each regime
    def getCone(self, p_regionByProcessor) :

        extrGrid = list(range(1))
        extrGrid[0][0] = p_regionByProcessor[0][0] - self.m_withdrawalRate
        extrGrid[0][1] = 1e30

        return extrGrid

    # permits to actualize the time (needed for simulation)
    def incrementActuTimeStep(self) :

        return None

    # defines a step in optimization
    # p_grid      grid at arrival step after command
    # p_stock     coordinate of the stock point to treat
    # p_condEsp   conditional expectation operator
    # p_phiIn     for each regime  gives the solution calculated at the previous step ( next time step by Dynamic Programming resolution)
    # for each regimes (column) gives the solution for each particle (row)
    def stepOptimize(self, p_grid, p_stock, p_condEsp, p_phiIn) :

        nbSimul = p_condEsp[0].getNbSimul()
        solutionAndControl = list(range(2))
        solutionAndControl[0] = np.zeros((nbSimul, 1))
        solutionAndControl[1] = np.zeros((nbSimul, 1))
        inflows = self.m_simulator.getParticles()[0,:].transpose()
        # level min of the stock
        minStorage = p_grid.getExtremeValues()[0][0]
        maxStorage = p_grid.getExtremeValues()[0][1]
        condExpSameStock = np.zeros(len(inflows))
        cashSameStock = np.zeros(len(inflows))

        for i in range(len(inflows)):
            stockWithInflows = np.zeros(1)
            stockWithInflows[0] = min(p_stock[0] + inflows[i], maxStorage)

            if p_grid.isInside(stockWithInflows):
                interpolatorCurrentStock = p_grid.createInterpolator(stockWithInflows)
                # cash flow at current stock and previous step
                cashSameStock[i] = interpolatorCurrentStock.apply(p_phiIn[0][i,:].transpose())
                # conditional expectation at current stock point
                condExpSameStock[i] = p_condEsp[0].getASimulation(i, interpolatorCurrentStock)

            else:
                cashSameStock[i] = - 1e30
                condExpSameStock[i] = - 1e30

        gainWithdrawal = np.zeros(len(inflows))
        cashWithdrawalStock = np.zeros(len(inflows))
        condExpWithdrawalStock  = np.zeros(len(inflows))
        withdrawalMax = np.zeros(nbSimul)

        for i in range(len(inflows)):
            stockWithIn = p_stock[0] + inflows[i]
            withdrawalMax[i] = min(stockWithIn - minStorage, self.m_withdrawalRate)

            if withdrawalMax[i] < (0. - 1e3 * 2.220446049250313e-16):
                cashWithdrawalStock[i] = - 1e30
                condExpWithdrawalStock[i] = - 1e30
                gainWithdrawal = - 1e30

            else:
                stockWithInflows = np.zeros(1)
                stockWithInflows[0] = min(stockWithIn - withdrawalMax[i], maxStorage)
                interpolatorWithdrawalStock = p_grid.createInterpolator(stockWithInflows)
                cashWithdrawalStock[i] = interpolatorWithdrawalStock.apply(p_phiIn[0][i,:].transpose())
                condExpWithdrawalStock[i] = p_condEsp[0].getASimulation(i, interpolatorWithdrawalStock)
                gainWithdrawal[i] = withdrawalMax[i]

        solutionAndControl[0][:,0] = cashSameStock
        solutionAndControl[1][:,0] = 0.
        espCondMax = condExpSameStock
        espCondWithdrawal = gainWithdrawal + condExpWithdrawalStock

        solutionAndControl[0][:,0] = np.where(espCondWithdrawal > espCondMax, gainWithdrawal + cashWithdrawalStock, solutionAndControl[0][:,0])
        solutionAndControl[1][:,0] = np.where(espCondWithdrawal > espCondMax, - withdrawalMax, solutionAndControl[1][:,0])

        return solutionAndControl

    # get number of regimes
    def getNbRegime(self) :

        return 1

    # number of controls
    def getNbControl(self):

        return 1

    # defines a step in simulation
    # Notice that this implementation is not optimal. In fact no interpolation is necessary for this asset.
    # This implementation is for test and example purpose
    # p_grid          grid at arrival step after command
    # p_continuation  defines the continuation operator for each regime
    # p_state         defines the state value (modified)
    # p_phiInOut      defines the value function (modified): size number of functions to follow
    def stepSimulate(self, p_grid, p_continuation, p_state, p_phiInOut) :

        ptStock = p_state.getPtStock()
        ptStockMax = ptStock
        # size of the stock
        maxStorage = p_grid.getExtremeValues()[0][1]
        # inflow
        inflow = p_state.getStochasticRealization()[0]
        # update storage
        ptStockCur = ptStock.copy()
        ptStockCur[0] += inflow
        ptStockCur[0] = min(ptStockCur[0], maxStorage) # deverse if necessary
        # if do nothing
        espCondMax = -1.e30

        if p_grid.isInside(ptStockCur):
            espCondMax = p_continuation[0].getValue(ptStockCur, p_state.getStochasticRealization())
            ptStockMax[0] += inflow

        # gain to add at current point
        phiAdd = 0.
        # if withdrawal
	# level min of the stock
        minStorage = p_grid.getExtremeValues()[0][0]
        withdrawalMax = min(p_state.getPtStock()[0] + inflow - minStorage, self.m_withdrawalRate)

        if (0. < withdrawalMax ):
            gainWithdrawal = withdrawalMax
            # level reached
            ptStockCur = ptStock + inflow - withdrawalMax
            ptStockCur[0] = min(ptStockCur[0], maxStorage)
            continuationWithdrawal = p_continuation[0].getValue(ptStockCur, p_state.getStochasticRealization())
            espCondWithdrawal = gainWithdrawal + continuationWithdrawal

            if espCondWithdrawal > espCondMax :
                espCondMax = espCondWithdrawal
                phiAdd = gainWithdrawal
                ptStockMax[0] = ptStockCur[0]

        # for return
        p_state.setPtStock(ptStockMax)
        p_phiInOut[0] += phiAdd

    # Defines a step in simulation using interpolation in controls
    # p_grid          grid at arrival step after command
    # p_control       defines the controls
    # p_state         defines the state value (modified)
    # p_phiInOut      defines the value function (modified): size number of functions to follow
    def stepSimulateControl(self, p_grid, p_control, p_state, p_phiInOut):

        ptStock = p_state.getPtStock()
        # size of the stock
        maxStorage = p_grid.getExtremeValues()[0][1]
        minStorage =  p_grid.getExtremeValues()[0][0]
        # inflow
        inflow = p_state.getStochasticRealization()[0]
        # get back control
        control = p_control[0].getValue(ptStock, p_state.getStochasticRealization())
        control = max(min(maxStorage - (ptStock[0] + inflow), control), minStorage - (ptStock[0] + inflow))
        ptStock[0] += inflow + control
        # for return
        p_state.setPtStock(ptStock)
        p_phiInOut[0] -= control

    # get size of the  function to follow in simulation
    def getSimuFuncSize(self):

        return 1

    # store the simulator
    def setSimulator(self, p_simulator):

        self.m_simulator = p_simulator

    # get the simulator back
    def getSimulator(self):

        return self.m_simulator



# Copyright (C) 2016 EDF, 2017, 2018  EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import os.path as osp
import sys
sys.path.append(osp.abspath(osp.dirname(osp.dirname(__file__))))
import numpy as np
import math
import random 
import unittest


# Ornstein Uhlenbeck simulator


# Ornstein Uhlenbeck simulator
class MeanRevertingSimulator :
    
    # Actualize trend
    def actualizeTrend(self) :
        
        self.m_trend = 0
        
        for i in list(range(len(self.m_sigma))) :
            self.m_trend += pow(self.m_sigma[i], 2.) / (2 * self.m_mr[i]) * (1 - math.exp(-2 * self.m_mr[i] * self.m_currentStep))
            
        self.m_trend *= 0.5
        
    # Constructor
    # p_curve  Initial forward curve
    # p_sigma  Volatility of each factor
    # p_mr     Mean reverting per factor
    # p_r      Interest rate
    # p_T      Maturity
    # p_nbStep Number of time step for simulation
    # p_nbSimul Number of simulations for the Monte Carlo
    # p_bForward true if the simulator is forward, false if the simulation is backward
    def __init__(self, p_curve, p_sigma, p_mr, p_r,p_T, p_nbStep, p_nbSimul, p_bForward) :
        
        self.m_curve = p_curve
        self.m_sigma = p_sigma
        self.m_mr = p_mr
        self.m_r = p_r
        self.m_T = p_T
        self.m_step = p_T / p_nbStep
        self.m_nbStep = p_nbStep
        self.m_nbSimul = p_nbSimul
        self.m_bForward = p_bForward
        self.m_currentStep = 0. if p_bForward else p_T
        self.m_OUProcess = np.zeros((len(p_sigma), p_nbSimul))

        np.random.seed(0)

        if self.m_bForward :
            self.m_OUProcess = np.zeros((len(p_sigma), p_nbSimul))
        else :
            #for i in list(range(self.m_OUProcess.shape[0])) :
            stDev = self.m_sigma * math.sqrt((1 - math.exp(-2 * self.m_mr * self.m_T)) / (2 * self.m_mr))
            self.m_OUProcess = stDev * np.random.randn(len(p_sigma), p_nbSimul)
            
        self.actualizeTrend()

    # a step forward for OU process
    def forwardStepForOU(self) :
        
        racine = math.sqrt((1 - np.exp(-2 * self.m_mr * self.m_step)) / (2 * self.m_mr))
        stDev = self.m_sigma * racine
        expActu = np.exp(-self.m_mr * self.m_step)
        normalSample = np.random.randn(len(self.m_sigma), self.m_nbSimul)
        increment = np.multiply(stDev, normalSample)
        # update OU process
        self.m_OUProcess = np.multiply(self.m_OUProcess, expActu) + increment
        
    # a step backward for OU process
    def backwardStepForOU(self) :
        
        if self.m_currentStep <= 0. :
            self.m_OUProcess = np.zeros((len(self.m_sigma), self.m_nbSimul))
        else :
            # use brownian bridge
            util = np.sinh(self.m_mr * self.m_currentStep) / np.sinh(self.m_mr * (self.m_currentStep + self.m_step))
            variance = pow(self.m_sigma, 2.) / (2* self.m_mr)*  ((1 - np.exp(-2 * self.m_mr * self.m_currentStep)) * pow(1 - np.exp(-self.m_mr * self.m_step)*util, 2.) + (1 - np.exp(-2 * self.m_mr * self.m_step))* pow(util, 2.))
            stdDev = np.sqrt(variance)
            self.m_OUProcess = self.m_OUProcess*util + np.einsum( "i,ij->ij",stdDev,np.random.randn(len(self.m_sigma), self.m_nbSimul))

        self.actualizeTrend()

    # get  current markov state
    def getParticles(self) :        
        return self.m_OUProcess

    # get one simulation
    # p_isim  simulation number
    # return the particle associated to p_isim
    # get  current markov state
    def getOneParticle(self, p_isim) :
        
        return self.m_OUProcess[:,p_isim]

    # a step forward for simulations
    def stepForward(self) :
        
        if self.m_bForward == False :
            pass
        else :
            self.m_currentStep += self.m_step
            self.actualizeTrend()
            self.forwardStepForOU()

    # return  the asset values (asset,simulations)
    def stepBackward(self) :
        
        if self.m_bForward == True :
            pass
        else :
            self.m_currentStep -= self.m_step
            self.actualizeTrend()
            self.backwardStepForOU()

    # a step forward for simulations
    # return  the asset values (asset,simulations)
    def stepForwardAndGetParticles(self) :
        
        if self.m_bForward == False :
            pass
        else :
            self.m_currentStep += self.m_step
            self.actualizeTrend()
            self.forwardStepForOU()
            
            return self.m_OUProcess

    # a step backward for simulations
    # return  the asset values (asset,simulations)
    def stepBackwardAndGetParticles(self) :
        
        if self.m_bForward == True :
            pass
        else :
            self.m_currentStep -= self.m_step
            self.actualizeTrend()
            self.backwardStepForOU()
            
            return self.m_OUProcess

    # From particles simulation for an  OU process, get spot price
    # p_particles  (dimension of the problem by number of simulations)
    # return spot price for all simulations
    def fromParticlesToSpot(self, p_particles) :
        
        values = np.zeros(p_particles.shape[1])
        curveCurrent = self.m_curve.get(self.m_currentStep)

        #for i in range(self.m_nbSimul) :
        values = np.multiply(curveCurrent, np.exp(-self.m_trend + np.sum(p_particles,axis=0)))

        return values

    # From one particle simulation for an  OU process, get spot price
    # p_oneParticle  One particle
    # return spot value
    def fromOneParticleToSpot(self, p_oneParticle) :
        
        curveCurrent = self.m_curve.get(self.m_currentStep)
        return curveCurrent * math.exp(-self.m_trend + np.sum(p_oneParticle))

    # get back asset spot value
    def getAssetValues(self) :
        
        return self.fromParticlesToSpot(self.m_OUProcess)

    # get back asset spot value
    # p_isim  simulation particle number
    # return spot value for this particle
    def getAssetValues2(self, p_isim) :
         
        return self.m_curve.get(self.m_currentStep) * math.exp(-self.m_trend + sum(self.m_OUProcess[:,p_isim]))

    # Get back attribute
    
    def getCurrentStep(self) :
        
        return self.m_currentStep

    def getT(self) :
        
        return self.m_T

    def getStep(self) :
        
        return self.m_step

    def getSigma(self) :
        
        return self.m_sigma

    def getMr(self) :
        
        return self.m_mr

    def getNbSimul(self) :
        
        return self.m_nbSimul

    def getNbSample(self) :
        
        return 1

    def getNbStep(self) :
        
        return self.m_nbStep

    def getDimension(self) :
        
        return len(self.m_sigma)

    # actualize at date t=0
    def getActu(self):
        return math.exp(- self.m_r* self.m_currentStep )

    # actualize on one step
    def getActuStep(self):
        return math.exp(- self.m_r* self.m_step )

    
    # forward or backward update
    # p_date  current date in simulator
    def updateDates(self, p_date) :
        
        if self.m_bForward :
            if p_date > 0. :
                self.stepForward()
        else :
            self.stepBackward()

    # forward or backward update for time
    def resetTime(self) :
        
        if self.m_bForward :
            self.m_currentStep = 0.
            self.m_OUProcess = 0.
        else :
            self.m_currentStep = self.m_T
            stDev = self.m_sigma * math.sqrt((1 - math.exp(-2 * self.m_mr * self.m_T)) / (2 * self.m_mr))
            self.m_OUProcess = stDev * np.random.randn(len(self.m_sigma), self.m_nbSimul)

        self.actualizeTrend()

    # update the number of simulations (forward only)
    # p_nbSimul  Number of simulations to update
    # p_nbSample Number of sample to update, useless here
    def updateSimulationNumberAndResetTime(self, p_nbSimul, p_nbSample) :
        
        if self.m_bForward == False :
            pass
        else :
            self.m_nbSimul = p_nbSimul
            self.m_OUProcess.reshape((len(self.m_sigma), p_nbSimul))
            self.m_currentStep = 0.
            self.m_OUProcess = 0.
            self.actualizeTrend()


class MeanRevertingSimulatorTest(unittest.TestCase):

    def test_callOption(self):
        
        nstep = 10
        timeGrid = odrsg.OneDimRegularSpaceGrid(0., 1. / nstep, nstep)
        futValues = np.zeros(nstep + 1) + 100.
        futureGrid = oddata.OneDimData(timeGrid, futValues)
        sigma = np.zeros(1) + 0.25
        mr = np.zeros(1) + 1.
        T = 2.
        nbStep = 10
        nbSimul = 1000000
        
        mrs = MeanRevertingSimulator(futureGrid, sigma, mr, T, nbStep, nbSimul, True)
        b = MeanRevertingSimulator(futureGrid, sigma, mr, T, nbStep, nbSimul, False)
        SF = np.zeros(nbStep / 2)
        SB = np.zeros(nbStep / 2)
        K = 100.
          
        def compareList(l, res) :
              
            for i in range(len(l)) :
                res[i] = max(l[i], 0.)
              
            return res
       
        for i in range(nbStep / 2) :
            particlesF = mrs.stepForwardAndGetParticles()
            spotF = mrs.fromParticlesToSpot(particlesF)
            lF = np.zeros(nbSimul)
            compareList(spotF - K, lF)
            SF[i] = np.mean(lF)
            particlesB = b.stepBackwardAndGetParticles()
            spotB = b.fromParticlesToSpot(particlesB)
            lB = np.zeros(nbSimul)
            compareList(spotB - K, lB)
            SB[i] = np.mean(lB)
            
        self.assertAlmostEqual(SF[nbStep/2-1], 7.069, None, None, 0.1)
        self.assertAlmostEqual(SB[nbStep/2-1], SF[nbStep/2-1], None, None, 0.2)
        
if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import StOptReg as reg
import StOptGrids
import StOptGlobal
import Simulators as sim
import Optimizers as opt
import Utils
import dp.DynamicProgrammingByRegressionHighLevel as dyn
import dp.SimulateRegressionControlHighLevel as srtc
import unittest
import importlib

accuracyClose = 1.5


# valorization of a given Lake on a  grid
# Gain are proportional to what is withdrawed from the storage
# Only inflows are stochastic
# p_grid             the grid
# p_maxLevelStorage  maximum level
# p_mesh             number of mesh
# p_bCheckClose      Do we check if optimization and simulations are close
def lake(p_grid, p_maxLevelStorage, p_mesh, p_bCheckClose):

     # test MPI
    moduleMpi4Py=importlib.util.find_spec('mpi4py')
    if (moduleMpi4Py is not None):
        from mpi4py import MPI

    # storage
    #########
    withdrawalRateStorage = 1000

    maturity = 1.
    nstep = 10

    # number of simulations
    nbsimulOpt = 8000

    # inflow model
    D0 = 50 # initial inflow
    m = D0 # average inflow
    sig = 5. # volatility
    mr = 5. # mean reverting
    # a backward simulator
    ######################
    bForward = False
    backSimulator = sim.AR1Simulator(D0, m, sig, mr, maturity, nstep, nbsimulOpt, bForward)
    # optimizer
    ##########
    storage = opt.OptimizeLakeAR1(withdrawalRateStorage)

    # regressor
    ###########
    nbMesh = np.array([], dtype = np.int32)

    if p_mesh > 0:
        nbMesh = np.zeros(1, dtype = np.int32) + p_mesh

    regressor = reg.LocalLinearRegression(nbMesh)
    # final value
    vFunction = Utils.ZeroPayOff()

    # initial values
    initialStock = np.zeros(1) + p_maxLevelStorage
    initialRegime = 0 # only one regime

    # Optimize
    fileToDump = "CondExpLakeHL"

    # link the simulations to the optimizer
    storage.setSimulator(backSimulator)
    valueOptim = dyn.DynamicProgrammingByRegressionHighLevel(p_grid, storage, regressor , vFunction, initialStock, initialRegime, fileToDump)

    nbsimulSim = 8000
    bForward = True

    forSimulator2 = sim.AR1Simulator(D0, m, sig, mr, maturity, nstep, nbsimulSim, bForward)

    storage.setSimulator(forSimulator2)
    valSimu2 = srtc.SimulateRegressionControl(p_grid, storage, vFunction, initialStock, initialRegime, fileToDump)

    print("valSimu2", valSimu2, "valueOptim", valueOptim)

class testLakeTest(unittest.TestCase):

    # linear interpolation
    def test_lakeLegendreLinear(self):

        # storage
        #########
        maxLevelStorage = 5000
        # grid
        ######
        nGrid = 10
        lowValues = np.zeros(1)
        step = np.zeros(1) + (maxLevelStorage / nGrid)
        nbStep = np.zeros(1, dtype = np.int32) + nGrid
        poly = np.zeros(1, dtype = np.int32) + 1
        grid = StOptGrids.RegularLegendreGrid(lowValues, step, nbStep, poly)
        nbmesh = 4

        lake(grid, maxLevelStorage, nbmesh, True)

    # quadratic interpolation on the basis functions
    def test_simpleStorageLegendreQuad(self):

        # storage
        #########
        maxLevelStorage = 5000
        # grid
        ######
        nGrid = 5
        lowValues = np.zeros(1)
        step = np.zeros(1) + (maxLevelStorage / nGrid)
        nbStep = np.zeros(1, dtype = np.int32) + nGrid
        poly = np.zeros(1, dtype = np.int32) + 2
        grid = StOptGrids.RegularLegendreGrid(lowValues, step, nbStep, poly)
        nbmesh = 4

        lake(grid, maxLevelStorage, nbmesh, True)

    # forget the AR1 model and suppose that inflows are iid
    def test_simpleStorageAverageInflows(self):

        # storage
        #########
        maxLevelStorage = 5000
        # grid
        ######
        nGrid = 10
        lowValues = np.zeros(1)
        step = np.zeros(1) + (maxLevelStorage / nGrid)
        nbStep = np.zeros(1, dtype = np.int32) + nGrid
        poly = np.zeros(1, dtype = np.int32) + 1
        grid = StOptGrids.RegularLegendreGrid(lowValues, step, nbStep, poly)
        nbmesh = 0

        lake(grid, maxLevelStorage, nbmesh, False)

if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2019 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import math
import StOptGrids
import StOptGlobal
import StOptGeners
import Simulators
import Optimizers
import Utils
import dp.DynamicProgrammingByTreeHighLevel as dyn
import dp.SimulateTreeControlHighLevel as simC
import dp.SimulateTreeHighLevel as simV
import unittest
import importlib

accuracyClose = 1.5


class ZeroFunction :
    
    def __init__(self) :
        
        return None

    def set(self, a, b, c) :
        
        return 0.

# valorization of a given gas storage on a  grid
# p_grid             the grid
# p_maxLevelStorage  maximum level
def gasStorage(p_grid, p_maxLevelStorage) :
    
    # storage
    injectionRateStorage = 60000
    withdrawalRateStorage = 45000
    injectionCostStorage = 0.35
    withdrawalCostStorage = 0.35

    maturity = 1.
    nstep = 100

    # define a a time grid
    timeGrid = StOptGrids.OneDimRegularSpaceGrid(0., maturity / nstep, nstep)
    # future values
    futValues = []

    # periodicity factor
    iPeriod = 52

    for i in list(range(nstep + 1)) :
        futValues.append(50. + 20. * math.sin((math.pi * i * iPeriod) / nstep))
      
    # define the future curve
    futureGrid = Utils.FutureCurve(timeGrid, futValues)
    
    # one dimensional factors
    mr = 0.29
    sigma =0.94

    # step
    dt = 1. / 100.
        
    # simulation dates
    dates =  dt * np.arange(0,nstep+1)

    # simulation dates
    tree = Simulators.TrinomialTreeOUSimulator(mr, sigma, dates)

    indexT = np.arange(0,nstep+1)
    tree.dump("Tree",indexT)

    # read archive
    archiveToRead = StOptGeners.BinaryFileArchive("Tree", "r");

    # backward simulator
    backSimulator = Simulators.MeanRevertingSimulatorTree(archiveToRead, futureGrid, sigma, mr)


    # optimizer
    storage = Optimizers.OptimizeGasStorageMeanRevertingTree(injectionRateStorage, withdrawalRateStorage, injectionCostStorage, withdrawalCostStorage)
    
    # final value
    vFunction = Utils.ZeroPayOff()
    
    # initial values
    initialStock = np.zeros(1) + p_maxLevelStorage
    initialRegime = 0 # only one regime
    
    # Optimize
    fileToDump = "CondExpGasStorage"
    
    # link the simulations to the optimizer
    storage.setSimulator(backSimulator)
    valueOptim = dyn.DynamicProgrammingByTreeHighLevel(p_grid, storage,  vFunction, initialStock, initialRegime, fileToDump)
    print("valueOptim", valueOptim)

    # new simulator for simulate optimal control
    nbsimulSim = 10000
    forwardSimulator = Simulators.MeanRevertingSimulatorTree(archiveToRead, futureGrid, sigma, mr,nbsimulSim)
     # link the simulations to the optimizer
    storage.setSimulator(forwardSimulator)

    # forward simulation of optimal control
    valueSim =simV.SimulateTreeHighLevel(p_grid,storage,vFunction, initialStock, initialRegime, fileToDump)
    print("Value simulation ", valueSim)

    # same with control
    forwardSimulator = Simulators.MeanRevertingSimulatorTree(archiveToRead, futureGrid, sigma, mr,nbsimulSim)
     # link the simulations to the optimizer
    storage.setSimulator(forwardSimulator)

    # forward simulation of optimal control
    valueSim =simC.SimulateTreeControlHighLevel(p_grid,storage,vFunction, initialStock, initialRegime, fileToDump)
    print("Value simulation with control", valueSim)
    
    return valueOptim



class testGasStorageTest(unittest.TestCase):
          
    def test_simpleStorage(self):
              
        # test MPI
        moduleMpi4Py=importlib.util.find_spec('mpi4py')
        if (moduleMpi4Py is not None):
            from mpi4py import MPI
        # storage
        #########
        maxLevelStorage = 90000
        # grid
        ######
        nGrid = 10
        lowValues = np.zeros(1)
        step = np.zeros(1) + (maxLevelStorage / nGrid)
        nbStep = np.zeros(1, dtype = np.int32) + nGrid
        grid = StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)
              
        gasStorage(grid, maxLevelStorage)
     

        
if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2018 The Regents of the University of California, Michael Ludkovski and Aditya Maheshwari
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)

from __future__ import division
import numpy as np
import math
import matplotlib.pyplot as plt
import time
import sys
from datetime import datetime
if (sys.version_info > (3, 0)):
        import pickle
else:
        import cPickle as pickle
       
import microgridDEA.parameters as bp
import microgridDEA.calEngineBatched as ce
import microgridDEA.forwardSimulations as fs



def runSimulation():

	# create an instance of parameter class.
	param = bp.basicVariables()

	# creates the list of objects defining the type of regression monte carlo.
	regParamObjcs = []
	for i in range(param.nstep):
		# for time steps from 0 to "param.designChangePnt-1", we use regress now-2D
		if i<param.designChangePnt:		
			regType = bp.regressionType('regress now 2D')
			regParamObjcs.append(regType)	
		else:
			# for time steps after "param.designChangePnt", we use grid-discretization
			regType = bp.regressionType('gd')
			regParamObjcs.append(regType)	

	# dump the regression parameters using pickle.
	name = param.regParamDump
	with open(name, 'wb') as output:
		pickle.dump(regParamObjcs, output, pickle.HIGHEST_PROTOCOL)

	ce.storageCalculation(param,regParamObjcs)

if __name__ == "__main__":

	runSimulation()	
	fs.forwardSimulations()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import math
import numpy as np
import unittest
import simulators.BlackScholesSimulator as bs
import utils.BasketOptions as bo
import StOptGrids 
import StOptReg as reg
import StOptGlobal
import Simulators as sim
import Optimizers as opt
import Utils
from utils.EuropeanOptions import CallOption
from utils.EuropeanOptions import PutOption as put
import importlib
import dp.DynamicProgrammingByRegressionDist as dynmpi

accuracyClose = 0.2
accuracyNearlyEqual = 0.05


# Analytical value
# p_S       asset value
# p_sigma   volatility
# p_r       interest rate
# p_strike  strike
# p_dates   possible exercise dates
# return option value
def analyticalValue(N, p_S, p_sigma, p_r, p_strike, p_dates) :
    
    analytical = 0.
    callOption = CallOption()
    
    for i in range(len(p_dates) - N, len(p_dates)) :
        analytical += callOption.operator(p_S, p_sigma, p_r, p_strike, p_dates[i])

    return analytical

# Classical resolution for swing
# p_sim             Monte Carlo simulator
# p_payOff          Option pay off
# p_regressor       regressor object
# p_dates           possible exercise dates
# p_N               number of exercises
def resolutionSwing(p_sim, p_payOff, p_regressor, p_dates, p_N) :
    
    if ((p_sim.getNbStep() + 1) == len(p_dates)) == False :
        pass
    
    # asset simulated under the neutral risk probability : get the trend of first asset to get interest rate
    # in this example the step between two exercises is given
    expRate = p_sim.getActuStep()
    
    # final payOff
    finalPayOff = p_payOff.applyVec(p_sim.getParticles())
    # Terminal function depending on simulations and stock already exercised
    cashNext = np.zeros((len(finalPayOff), p_N))

    for i in range(len(finalPayOff)) :
        cashNext[i,:] = (finalPayOff[i]).transpose()

    cashPrev = np.zeros((len(finalPayOff), p_N))

    for iStep in range(len(p_dates) - 1)[::-1] :
        asset = p_sim.stepBackwardAndGetParticles()
        payOffLoc = p_payOff.applyVec(asset)
        payOffLoc = payOffLoc.squeeze()

        # conditional expectation
        if iStep == 0 :
            p_regressor.updateSimulations(True, asset)
        else :
            p_regressor.updateSimulations(False, asset)

        # store conditional expectations
        vecCondEspec = list(range(p_N))

        for iStock in range(p_N):
            vecCondEspec[iStock] = p_regressor.getAllSimulations(cashNext[:,iStock]).squeeze() * expRate

        # arbitrage
        for iStock in range(p_N - 1):
            cashPrev[:,iStock] = np.where(payOffLoc + vecCondEspec[iStock + 1] > vecCondEspec[iStock], payOffLoc + expRate * cashNext[:,iStock + 1], expRate * cashNext[:,iStock])
            
        # last stock
        cashPrev[:,p_N - 1] = np.where(payOffLoc > vecCondEspec[p_N - 1], payOffLoc, expRate * cashNext[:,p_N - 1])
            
        # switch
        tempVec = cashNext
        cashNext = cashPrev
        cashPrev = tempVec

    return np.mean(cashNext[:,0])

def resolutionSwingContinuation(p_sim, p_payOff, p_regressor, p_dates, p_N):
    
    if (p_sim.getNbStep() + 1 == len(p_dates)) == False:
        pass
    
    # asset simulated under the neutral risk probability : get the trend of first asset to get interest rate
    # in this example the step between two exercises is given
    expRate = math.exp(-p_sim.getStep() * p_sim.getMu()[0])
    lowValues = np.zeros(1)
    step = np.zeros(1)
    lowValues[0] = 0.
    step[0] = 1
    nbStep = np.zeros(1, dtype = np.int32)
    nbStep[0] = p_N - 1
    regular = StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)
    extremeGrid = np.zeros(2)
    extremeGrid[0] = lowValues
    extremeGrid[1] = lowValues + step * nbStep
    
    # final payOff
    finalPayOff = p_payOff.applyVec(p_sim.getParticles())
    # Terminal function depending on simulations and stock already exercised
    cashNext = np.zeros((len(finalPayOff), regular.getNbPoints()))

    for i in range(len(finalPayOff)):
        cashNext[i,:] = (np.zeros(regular.getNbPoints()) + finalPayOff[i]).transpose()

    cashPrev = np.zeros((len(finalPayOff), regular.getNbPoints()))

    for iStep in range(len(p_dates) - 2)[::-1] :
        asset = p_sim.stepBackwardAndGetParticles()
        payOffLoc = p_payOff.applyVec(asset)

        # conditional expectation
        if iStep == 0 :
            p_regressor.updateSimulations(True, asset)
        else :
            p_regressor.updateSimulations(False, asset)
            
        # continuation value object dealing with stocks
        continuation = reg.ContinuationValue(regular, p_regressor, cashNext)
        # iterator on grid points
        iterOnGrid = regular.getGridIterator()
        
        while iterOnGrid.isValid():
            CoordStock = iterOnGrid.getCoordinate()
            # use continuation to get realization of condition expectation
            conditionExpecCur = np.multiply(expRate, continuation.getAllSimulations(CoordStock).squeeze())
            
            if CoordStock[0] + 1 <= extremeGrid[1] :
                conditionExpecNext = np.multiply(expRate, continuation.getAllSimulations(CoordStock + 1).squeeze())
                cashPrev[:,iterOnGrid.getCount()] = np.where(payOffLoc + conditionExpecNext > conditionExpecCur, payOffLoc + np.multiply(expRate, cashNext[:,iterOnGrid.getCount() + 1]), np.multiply(expRate, cashNext[:,iterOnGrid.getCount()]))
            else :
                cashPrev[:,iterOnGrid.getCount()] = np.where(payOffLoc > conditionExpecCur, payOffLoc, np.multiply(expRate, cashNext[:,iterOnGrid.getCount()]))
                
            iterOnGrid.next()
            
        # switch pointer
        tempVec = cashNext
        cashNext = cashPrev
        cashPrev = tempVec
        
    return np.mean(cashNext[:,0])

#  function to test stock in dimension above 1
def multiStock(p_ndim):
    
    moduleMpi4Py=importlib.util.find_spec('mpi4py')
    if (moduleMpi4Py is not None):
        from mpi4py import MPI
        world = MPI.COMM_WORLD
        initialValues = np.zeros(1) + 1.
        sigma = np.zeros(1) + 0.2
        mu = np.zeros(1) + 0.05
        corr = np.ones((1,1))
        # number of step
        nStep = 20
        # exercice date
        dates = np.linspace(0., 1., nStep + 1)
        # exercice dates
        N = 3
        T = 1.
        strike = 1.
        nbSimul = 10000
        nMesh = 4
        # payoff
        payoff = Utils.BasketCall(strike)
        # store sequential
        valueSeq = 0
        # mesh
        nbMesh = np.zeros(1, dtype = np.int32) + nMesh
        
        if world.rank == 0:
            # simulator
            simulator = bs.BlackScholesSimulator(initialValues, sigma, mu, corr, dates[len(dates) - 1], len(dates) - 1, nbSimul, False)
            # regressor
            regressor = reg.LocalLinearRegression(nbMesh)
            # bermudean value
            valueSeq = resolutionSwing(simulator, payoff, regressor, dates, N)
            
        # simulator
        simulator = sim.BlackScholesSimulator(initialValues, sigma, mu, corr, dates[len(dates) - 1], len(dates) - 1, nbSimul, False)
        # grid
        lowValues = np.zeros(p_ndim)
        step = np.zeros(p_ndim) + 1.
        # the stock is discretized with values from 0 to N included
        nbStep = np.zeros(p_ndim, dtype = np.int32) + N
        grid = StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)
        # final value
        vFunction = Utils.PayOffFictitiousSwing(payoff, N)
        # optimizer
        optimizer = opt.OptimizerFictitiousSwingBlackScholes(payoff, N, p_ndim)
        # initial values
        initialStock = np.zeros(p_ndim)
        initialRegime = 0
        fileToDump = "CondExpSwing"
        # regressor
        regressor = reg.LocalLinearRegression(nbMesh)
        bOneFile = False
        # link the simulations to the optimizer
        optimizer.setSimulator(simulator)
        valueParal = dynmpi.DynamicProgrammingByRegressionDist(grid, optimizer, regressor, vFunction, initialStock, initialRegime, fileToDump, bOneFile)
        return valueParal, valueSeq

class testSwingOptionTest(unittest.TestCase) :
     
    def test_swingOptionInOptimization(self):
         
        initialValues = np.zeros(1) + 1.
        sigma = np.zeros(1) + 0.2
        mu = np.zeros(1) + 0.05
        corr = np.ones((1,1))
        # number of step
        nStep = 30
        # exercise date
        dates = np.linspace(0., 1., nStep + 1)
        N = 3 # 3 exercise dates
        T = 1.
        strike = 1.
        nbSimul = 200000
        nMesh = 16
        # payoff
        payoff = bo.BasketCall(strike)
        # analytical
        analytical = analyticalValue(N, initialValues[0], sigma[0], mu[0], strike, dates)
        
        # mesh
        nbMesh = np.zeros(1, dtype = np.int32) + nMesh
        # simulator
        simulator = bs.BlackScholesSimulator(initialValues, sigma, mu, corr, dates[len(dates) - 1], len(dates) - 1, nbSimul, False)
        # regressor
        regressor = reg.LocalLinearRegression(nbMesh)
        # bermudean value
        valueSeq = resolutionSwing(simulator, payoff, regressor, dates, N)
        
        # simulator
        simulator = bs.BlackScholesSimulator(initialValues, sigma, mu, corr, dates[len(dates) - 1], len(dates) - 1, nbSimul, False)
        # regressor
        regressor = reg.LocalLinearRegression(nbMesh)
        # using continuation values
        valueSeqContinuation = resolutionSwingContinuation(simulator, payoff, regressor, dates, N)
        print("valueSeq", valueSeq, "valueSeqContinuation", valueSeqContinuation)
        
        self.assertAlmostEqual(valueSeq, analytical, None, None, accuracyClose)
        
    def test_swingOption2D(self):
        
        moduleMpi4Py=importlib.util.find_spec('mpi4py')
        if (moduleMpi4Py is not None):
            from mpi4py import MPI
            world = MPI.COMM_WORLD
            val = multiStock(2)
            
            if world.rank == 0:
                self.assertAlmostEqual(val[1] * 2, val[0], None, None, accuracyNearlyEqual)
                
    def test_swingOption3D(self):
        
        moduleMpi4Py=importlib.util.find_spec('mpi4py')
        if (moduleMpi4Py is not None):
            from mpi4py import MPI
            world = MPI.COMM_WORLD
            val = multiStock(3)
            
            if world.rank == 0:
                self.assertAlmostEqual(val[1] * 3, val[0], None, None, accuracyNearlyEqual)
                
if __name__ == '__main__':
    
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import StOptGrids 
import StOptReg
import StOptGlobal
import StOptGeners
import imp
import numpy as np


def DynamicProgrammingByRegressionDist(p_grid, p_optimize, p_regressor, p_funcFinalValue, p_pointStock, p_initialRegime, p_fileToDump, p_bOneFile) :
    
    try:
        imp.find_module('mpi4py')
        found = True
    except:
        print("Not parallel module found ")
        found = False
        
    if found :
        # from the optimizer get back the simulation
        simulator = p_optimize.getSimulator()
        fin = StOptGlobal.FinalStepDPDist(p_grid, p_optimize.getNbRegime(),p_optimize.getDimensionToSplit())
        # final values
        valuesNext = fin.set(p_funcFinalValue, np.asarray(simulator.getParticles()))
        
        from mpi4py import MPI
        world = MPI.COMM_WORLD
        # dump
        toDump = p_fileToDump
        
        # test if one file generated
        if p_bOneFile is False:
            toDump += "_" + str(world.rank)
            
        ar = StOptGeners.BinaryFileArchive(toDump, "w")
        # name for object in archive
        nameAr = "Continuation"
        
        nsteps =simulator.getNbStep()
        # iterate on time steps
        for iStep in range(nsteps) :
            asset = simulator.stepBackwardAndGetParticles()
            
            # conditional espectation operator
            if iStep == (simulator.getNbStep() - 1):
                p_regressor.updateSimulations(True, asset)
            
            else:
                p_regressor.updateSimulations(False, asset)
            
            # transition object
            transStep = StOptGlobal.TransitionStepRegressionDPDist(p_grid, p_grid, p_optimize)
            valuesAndControl = transStep.oneStep(valuesNext, p_regressor)
            transStep.dumpContinuationValues(ar, nameAr,  nsteps - 1 - iStep, valuesNext, valuesAndControl[1], p_regressor, p_bOneFile)
            valuesNext = valuesAndControl[0]
            
        # reconstruct a small grid for interpolation
        return StOptGlobal.reconstructProc0Mpi(p_pointStock, p_grid, valuesNext[p_initialRegime],p_optimize.getDimensionToSplit()).mean()



# Copyright (C) 2019 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import unittest
import random
import StOptGrids
import StOptReg
import StOptTree
import StOptGlobal
import Simulators
import StOptGeners


# unit test for tree

class testTree(unittest.TestCase):

    def testContinuation(self):

        # Mean Reverting model
        mr = 0.3;
        sig = 0.6;

        # nb grid points
        nbStock=4
        
        # step
        dt = 1. / 100.
        
        # simulation dates
        dates =  dt * np.arange(0,16)

        # simulaton dates
        tree = Simulators.TrinomialTreeOUSimulator(mr, sig, dates)

        iFirst = 10
        iLast = 14

        # nodes at dates 5
        points = tree.getPoints(iFirst)

        # nodes at last date
        pointsNext = tree.getPoints(iLast)

        # probabilities
        proba = tree.getProbability(iFirst, iLast)

        # connection matrix
        connectAndProba =  tree.calConnected(proba);

        # to regress
        toTreeress= np.zeros( (nbStock, np.shape(pointsNext)[1]))
        for i in range(nbStock):
            toTreeress[i,:] = i + pointsNext[0,:]

        # grid for storage
        grid = StOptGrids.RegularSpaceGrid(np.array([0.]), np.array([1.]), np.array([nbStock - 1]))

        # conditional expectation object by trees
        tree= StOptTree.Tree(connectAndProba[1],connectAndProba[0])

        # conditional expectation taken
        valTree = tree.expCondMultiple(toTreeress)

        # continuation object
        continuation = StOptTree.ContinuationValueTree(grid, tree,  toTreeress.transpose())

        # interpolation point
        ptStock = np.array([0.5*nbStock])

        #  conditional expectation using continuation object
        treeByContinuation = continuation.getValueAtNodes(ptStock);

        for i in range(np.shape(points)[1]):
            self.assertAlmostEqual(treeByContinuation[i],valTree[int(nbStock / 2), i], 7,"Difference between function and its condition expectation estimated greater than tolerance")
 
        # into a list  : two times to test 2 regimes
        listToReg = []
        listToReg.append(toTreeress.transpose())
        listToReg.append(toTreeress.transpose())
 
        #default non compression
        archiveToWrite = StOptGeners.BinaryFileArchive("MyArchive", "w");
        # continuation value dump
        archiveToWrite.dumpGridTreeValue("toStore",1,listToReg,tree,grid)

        # read archive
        archiveToRead =  StOptGeners.BinaryFileArchive("MyArchive","r")
        contValues = archiveToRead.readGridTreeValue(1,"toStore")
        # interpolate
        value =contValues[0].getValues(ptStock)
        for i in range(np.shape(points)[1]):
            self.assertAlmostEqual(value[i],valTree[int(nbStock / 2), i], 7,"Difference between function and its condition expectation estimated greater than tolerance")
  
if __name__ == '__main__': 
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
# Final value function for swings in multidimensionnal case


# final function payoff for a fictitious swing
class FinalValueFictitiousFunction:
    
    # Constructor
    def __init__(self, p_pay, p_nExerc):
        
        self.m_pay = p_pay
        self.m_nExerc = p_nExerc
        
    # final function
    # p_ireg   regime number
    # p_stock  position in the stock
    # p_state  position in the stochastic state
    def set(self, ireg, p_stock, p_state):
        
        idec = len(p_stock) * self.m_nExerc - p_stock.sum()
        
        return min(idec, len(p_stock)) * self.m_pay.apply(p_state)



# Copyright (C) 2018 The Regents of the University of California, Michael Ludkovski and Aditya Maheshwari
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)

from __future__ import division
import numpy as np
import math
import matplotlib.pyplot as plt
import time

import StOptGrids
import StOptReg
import StOptGeners

import microGrid.simulateState as simX
import microGrid.valueNext as fv


# returns the value function at the final step. Depending upon the type of RMC method, the size of the matrix may be different.
def finalStepValue(param,demand,inventory):

	return fv.finalValue(param,demand,inventory)


def createXMatrix(param,asset, inventory=[]):

	if param.rmctype == 'regress now 2D':
		xMatrix = np.zeros((2,len(asset)))
		xMatrix[0,:] = asset
		xMatrix[1,:] = inventory
		return xMatrix

	if param.rmctype == 'gd':
		xMatrix = np.zeros((1,len(asset)))
		xMatrix[0,:] = asset
		return xMatrix



def storageCalculation(param):

	# for regress now 2D, a residual demand process and fixed inventory vector at every time step
	# for gd,  a price process and fixed inventory grid at every time step 
	X0 = np.linspace(-param.ampl,param.ampl,param.nbsimulOpt)
	residualDemand,inventory = simX.residualDemand(param, param.nstep, param.nbsimulOpt, param.maturity, X0, param.rmctype)
	inventory[0] = 0

	if param.rmctype == 'gd':

		stepI = param.I_minMax[1]/param.meshI										
		lowValues =np.array([param.I_minMax[0] , 0.],dtype=np.float64)				
		step = np.array([stepI , 1],dtype=np.float64)  								
		nbStep = np.array([param.meshI , param.H], dtype=np.int32)					
		grid = StOptGrids.RegularSpaceGrid(lowValues,step,nbStep)		

	if param.rmctype == 'regress now 2D':
		
		lowValues =np.array([0.],dtype=np.float64)				            # low value for the mesh
		step = np.array([1],dtype=np.float64)  								# size of the mesh
		nbStep = np.array([param.H], dtype=np.int32)						# number of step
		grid = StOptGrids.RegularSpaceGrid(lowValues,step,nbStep)	

	# Value at the final step. 
	valueNext = finalStepValue(param,residualDemand[:,-1],inventory)

	archiveToWrite = StOptGeners.BinaryFileArchive(param.filetoDump,"w")


	# iterate on time steps
	for iSteps in range(param.nstep-1,-1,-1):

		print("iSteps", iSteps )
		demand = residualDemand[:,iSteps]

		# rmc type and regression method type. The choice is made in the parameter class. 
		if param.rmctype == 'gd':
			nbMesh = np.array([param.meshX],dtype=np.int32)
			regressor = StOptReg.LocalLinearRegression(False,createXMatrix(param,demand),nbMesh)

		elif param.rmctype == 'regress now 2D':
			if param.regType == 'piecewiseLinear':
				nbMesh = np.array([param.meshX, param.meshI],dtype=np.int32)
				regressor = StOptReg.LocalLinearRegression(False,createXMatrix(param,demand, inventory),nbMesh)

			if param.regType == 'globalPolynomial':
				regressor = StOptReg.GlobalCanonicalRegression(False,createXMatrix(param,demand, inventory),param.degree)

			if param.regType == 'kernel':
				regressor = StOptReg.LocalGridKernelRegression(False,createXMatrix(param,demand, inventory),param.bandwidth,param.factPoint, True)

		# Dump the continuation values in the archive:
		archiveToWrite.dumpGridAndRegressedValue("toStore", param.nstep - iSteps, [valueNext], regressor,grid)

		# Read the regressed values
		archiveToRead =  StOptGeners.BinaryFileArchive(param.filetoDump,"r")
		contValues = archiveToRead.readGridAndRegressedValue(param.nstep - iSteps,"toStore")


		valueNext, control_d, StOutput = fv.optimization(contValues[0], demand, inventory, param)


		#######################################################################################################
		# uncomment below to observe following graphs:
		# 1) Cost as a function of residual demand at differnet levels of the inventory.
		# 2) Cost as a function of inventory at differnet levels of the residual demand
		# 3) Control strategy when the generator is ON and when it is OFF.
		# if (iSteps == 0):

		# 	demandCheck = demand #np.tile(demand,param.meshI)
		# 	inventoryCheck = inventory #np.repeat(inventory,param.meshX)

		#  	for q in range(2):
		#  		# plt.figure(q+1)
		#  		plt.scatter(demandCheck,valueNext[:,q],c=inventoryCheck, lw=0, s=10)
		#  		plt.xlabel("demand")
		#  		plt.ylabel("cost")
		#  		name = "q="+str(q)
		#  		plt.title(name)
		#  		plt.colorbar()
		#  		filename = 'q'+str(q)+'_iSteps'+str(iSteps)+'_costDemand.png'
		#  		plt.savefig(filename)
		#  		plt.close()


		#  	for q in range(2):
		#  		# plt.figure(q+1)
		#  		plt.scatter(inventoryCheck,valueNext[:,q],c=demandCheck, lw=0, s=10)
		#  		plt.xlabel("inventory")
		#  		plt.ylabel("cost")
		#  		#plt.ylim(0,1000)
		#  		name = "q="+str(q)
		#  		plt.title(name)
		#  		plt.colorbar()
		#  		filename = 'q'+str(q)+'_iSteps'+str(iSteps)+'_costinventory.png'
		#  		plt.savefig(filename)
		#  		plt.close()

		#  	for q in range(2):
		#  		plt.scatter(demandCheck,inventoryCheck,c=control_d[:,q],lw=0, s=10)
		#  		plt.xlabel('demand')
		#  		plt.ylabel('inventory')
		#  		name = "q="+str(q)
		#  		plt.title(name)
		#  		plt.colorbar()
		#  		filename = 'q'+str(q)+'_iSteps'+str(iSteps)+'_controlMap.png'
		#  		plt.savefig(filename)
		#  		plt.close()




import glob
import sys
import string
import os


# dico  de fichier
listFile  = {}
# liste modif include
listInclude = {}

pathh = ['../StOpt', '../test']
for xpath in pathh:
    for path, subdirs, files in os.walk(xpath):
        for name in files:
            print " path", path , "/" , name.rstrip()
            fp =open(path+"/"+name.rstrip(),'r')   
            dlines = fp.readlines()    
            fp.close()
            fp =open(path+"/"+name.rstrip(),'w')  
            for l in dlines:
                if (l.find("Eclipse Public License")):
                    l= string.replace(l,"Eclipse Public License","GNU Lesser General Public License (GNU LGPL)")
                fp.write(l)
            fp.close()

"""                   
                if (l.find("2015") !=-1):      
                 l = string.replace(l,"2015","2016") 
                if (l.find("FiME") !=-1):
                 l = string.replace(l,"FiME","EDF")
                if (l.find("FIME") !=-1):
                 l = string.replace(l,"FIME","EDF")
"""



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import math as maths
import utils.BasketOptions as bo
import simulators.BlackScholesSimulator as bs
import utils.americanOption as amer
import StOptGrids
import StOptReg as reg
import unittest



def american(p_nDim, p_nbSimul, p_nMesh, p_bRotate) :
    
    initialValues = np.zeros(p_nDim) + 1.0
    sigma = np.zeros(p_nDim) + 0.2
    mu = np.zeros(p_nDim) + 0.05
    corr = np.zeros((p_nDim,p_nDim))
    T = 1.
    nDate = 10
    np.fill_diagonal(corr, 1.)
    strike = 1.

    # simulator
    simulator = bs.BlackScholesSimulator(initialValues, sigma, mu, corr, T, nDate, p_nbSimul, False)
    # payoff
    payoff = bo.BasketPut(strike)
    # mesh
    nbMesh = np.zeros(p_nDim, dtype=np.int32) + p_nMesh
    # regressor
    regressor = reg.LocalLinearRegression(nbMesh,p_bRotate)
    # bermudean value
    value = amer.resolution(simulator, payoff, regressor)
    return value

# test cases
class testAmericanOptionTest(unittest.TestCase):
    
    def test_americanBasket1D(self):
  
        # dimension
        nDim = 1
        nbSimul = 500000
        nbMesh = 16
        referenceValue = 0.06031
        accuracyEqual = 0.13
               
        self.assertAlmostEqual(american(nDim, nbSimul, nbMesh, False), referenceValue, None, None, accuracyEqual)
        
    def test_americanBasket2D(self):    
        
        # dimension
        nDim = 2
        nbSimul = 1000000
        nbMesh = 16
        referenceValue = 0.03882
        accuracyEqual = 0.4
            
        self.assertAlmostEqual(american(nDim, nbSimul, nbMesh, False), referenceValue, None, None, accuracyEqual)
        self.assertAlmostEqual(american(nDim, nbSimul, nbMesh, True), referenceValue, None, None, accuracyEqual)
     
    def test_americanBasket3D(self):
    
        # dimension
        nDim = 3
        nbSimul = 2000000
        nbMesh = 10
        referenceValue = 0.02947
        accuracyEqual = 0.5
           
        self.assertAlmostEqual(american(nDim, nbSimul, nbMesh,False), referenceValue, None, None, accuracyEqual)
        self.assertAlmostEqual(american(nDim, nbSimul, nbMesh,True), referenceValue, None, None, accuracyEqual)
      
    def test_americanBasket4D(self):
        
        # dimension
        nDim = 4
        nbSimul = 2000000
        nbMesh = 6
        referenceValue = 0.02404
        accuracyEqual = 1.
           
        self.assertAlmostEqual(american(nDim, nbSimul, nbMesh,False), referenceValue, None, None, accuracyEqual)

if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import StOptReg as reg
import StOptGrids
import StOptGeners
from StOptGlobal import StateWithStocks


def SimulateRegressionControl(p_grid, p_optimize, p_funcFinalValue, p_pointStock,
                              p_initialRegime, p_fileToDump, key="Continuation"):
    """
    Simulate the optimal strategy.

    Parameters:
    -----------
    p_grid
        Grid used for  deterministic state (stocks for example).
    p_optimize
        Optimizer defining the optimization between two time steps.
    p_funcFinalValue
        Function defining the final value.
    p_pointStock
        Initial point stock.
    p_initialRegime
        Regime at initial date.
    p_fileToDump
        Name of the file used to dump continuation values in optimization.
    """
    simulator = p_optimize.getSimulator()
    nsteps = simulator.getNbStep()
    nsims = simulator.getNbSimul()
    dim = simulator.getDimension()
    particle0 =  simulator.getParticles()[:,0]
    states = [StateWithStocks(p_initialRegime, p_pointStock, particle0)
              for _ in range(nsims)]
    # Retrieve the file containing the continuation values:
    ar = StOptGeners.BinaryFileArchive(p_fileToDump, "r")
    # Cost function
    costFunction = np.zeros((p_optimize.getSimuFuncSize(), nsims))
    # Iterate on time steps.
    for istep in range(nsteps) :
        continuation = ar.readGridAndRegressedValue(istep, key)
        grid = continuation[0].getGrid()
        for i in range(nsims):
            state = states[i]
            p_optimize.stepSimulate(grid, continuation, state,
                                    costFunction[:, i])
        particles = simulator.stepForwardAndGetParticles()
        for i in range(nsims) :
            states[i].setStochasticRealization(particles[:, i])

    # Final step: accept to exercise if not already done entirely.
    for i in range(simulator.getNbSimul()):
        costFunction[0, i] += p_funcFinalValue.set(states[i].getRegime(),
                                                   states[i].getPtStock(),
                                                   states[i].getStochasticRealization())

    # Average gain/cost.
    return costFunction.mean()







# Copyright (C) 2019 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import math
import StOptGrids
import StOptTree
import StOptGlobal
import StOptGeners
import Utils
import Simulators
import unittest

# test cases
class testAmericanOptionTest(unittest.TestCase):
    
    def test_americanTree(self):
  
        # define a a time grid
        timeGrid = StOptGrids.OneDimRegularSpaceGrid(0., 2., 1)
        # future values
        futureGrid =  Utils.FutureCurve(timeGrid, [50.,50.])

        # degenerate trinomial tri
        mr = 0.00001
        sigma =0.2

        # interest rate
        r = 0.05

        # number of time steps
        nstep =50
        
        # maturity
        T =1.
        
        # step
        dt = T / nstep
        
        # simulation dates
        dates =  dt * np.arange(0,nstep+1)

        # simulation dates
        treeSim = Simulators.TrinomialTreeOUSimulator(mr, sigma, dates)

        indexT = np.arange(0,nstep+1, 5)
        treeSim.dump("Tree",indexT)

        # read archive
        archiveToRead = StOptGeners.BinaryFileArchive("Tree", "r");
        
       # FIRST VALORIZATION
        ####################
        
        # backward simulator
        backSimulator = Simulators.MeanRevertingSimulatorTree(archiveToRead, futureGrid, sigma, mr)

        # strike
        strike  = 50.

        # actu
        actu = np.exp(r*dates[indexT[-1]])
        # spot : add interest rate
        spot = backSimulator.getSpotValues()*actu
        # actualized payoff
        val1= np.where( strike-spot>0,strike-spot,0)/actu
        for istep in np.arange(np.shape(indexT)[0]-1):
            # backward in simulator
            backSimulator.stepBackward()
            # get back probability matrix
            proba = backSimulator.getProba()
            # and connection matrix
            connected = backSimulator.getConnected()
            # creta tree for conditional expectation
            tree=StOptTree.Tree(proba,connected)      
            # interest rates
            actu = np.exp(r*dates[indexT[-2-istep]])
            # spot : add interest rate
            spot = backSimulator.getSpotValues()*actu
            # pay off 
            payOff = np.where( strike-spot>0,strike-spot,0)/actu
            # actualize value
            val1 = tree.expCond(val1)
            # arbitrage
            val1 = np.where( val1> payOff, val1, payOff)

        final = val1[0]
        
        indexT = np.arange(0,nstep+1)
        treeSim.dump("Tree",indexT)

        # read archive
        archiveToRead = StOptGeners.BinaryFileArchive("Tree", "r");
        
        # backward simulator
        backSimulator = Simulators.MeanRevertingSimulatorTree(archiveToRead, futureGrid, sigma, mr)

        # SECOND VALORIZATION
        ####################
        # actu
        actu = np.exp(r*dates[-1])
        # spot : add interest rate
        spot = backSimulator.getSpotValues()*actu
        # actualized payoff
        val2= np.where( strike-spot>0,strike-spot,0)/actu

        for istep in np.arange(nstep):
            # backward in simulator
            backSimulator.stepBackward()
            # get back probability matrix
            proba = backSimulator.getProba()
            # and connection matrix
            connected = backSimulator.getConnected()
            # creta tree for conditional expectation
            tree=StOptTree.Tree(proba,connected)      
            # actualize value
            val2 = tree.expCond(val2)

            if ((istep+1)%5==0):
                # interest rates
                actu = np.exp(r*dates[-2-istep])
                # spot : add interest rate
                spot = backSimulator.getSpotValues()*actu
                # pay off 
                payOff = np.where( strike-spot>0,strike-spot,0)/actu
                # arbitrage
                val2 = np.where( val2> payOff, val2, payOff)

        print("val1 " , val1 , " val2 " , val2)
        self.assertAlmostEqual(val1, val2, None, None, 0.000001)
        
 
if __name__ == '__main__':
    unittest.main()
